from .arc import Arc
from .box import Box
from .circle import Circle
from .control_point import ControlPoint
from .curve import Curve
from .ellipse import Ellipse
from .line import Line
from .mesh import Mesh
from .plane import Plane
from .point import Point
from .point_cloud import PointCloud
from .polycurve import Polycurve
from .polyline import Polyline
from .region import Region
from .spiral import Spiral
from .surface import Surface
from .vector import Vector

# re-export them at the geometry package level
__all__ = [
    "Arc",
    "Line",
    "Mesh",
    "Plane",
    "Point",
    "Polyline",
    "Region",
    "Vector",
    "Box",
    "Circle",
    "ControlPoint",
    "Ellipse",
    "PointCloud",
    "Polycurve",
    "Spiral",
    "Surface",
    "Curve",
]



from enum import Enum
from typing import List, Optional

from specklepy.objects.base import Base
from specklepy.objects.geometry import Point

CHUNKABLE_PROPS = {
    "vertices": 100,
    "faces": 100,
    "colors": 100,
    "textureCoordinates": 100,
    "test_bases": 10,
}

DETACHABLE = {"detach_this", "origin", "detached_list"}


class FakeGeo(
    Base, speckle_type="FakeGeo", chunkable={"dots": 50}, detachable={"pointslist"}
):
    pointslist: Optional[List[Base]] = None
    dots: Optional[List[int]] = None


class FakeDirection(Enum):
    NORTH = 1
    EAST = 2
    SOUTH = 3
    WEST = 4


class FakeMesh(
    FakeGeo, speckle_type="FakeMesh", chunkable=CHUNKABLE_PROPS, detachable=DETACHABLE
):
    vertices: Optional[List[float]] = None
    faces: Optional[List[int]] = None
    colors: Optional[List[int]] = None
    textureCoordinates: Optional[List[float]] = None
    cardinal_dir: Optional[FakeDirection] = None
    test_bases: Optional[List[Base]] = None
    detach_this: Optional[Base] = None
    detached_list: Optional[List[Base]] = None
    _origin: Optional[Point] = None

    # def __init__(self, **kwargs) -> None:
    #     super(FakeMesh, self).__init__(**kwargs)
    #     self.add_chunkable_attrs(**CHUNKABLE_PROPS)
    #     self.add_detachable_attrs(DETACHABLE)

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value: Point):
        self._origin = value



from dataclasses import dataclass, field
from typing import List, Tuple

from specklepy.objects.base import Base
from specklepy.objects.geometry.point import Point
from specklepy.objects.interfaces import IHasArea, IHasUnits, IHasVolume


@dataclass(kw_only=True)
class Mesh(
    Base,
    IHasArea,
    IHasVolume,
    IHasUnits,
    speckle_type="Objects.Geometry.Mesh",
    detachable={"vertices", "faces", "colors", "textureCoordinates", "vertexNormals"},
    chunkable={
        "vertices": 31250,
        "faces": 62500,
        "colors": 62500,
        "textureCoordinates": 31250,
        "vertexNormals": 31250,
    },
    serialize_ignore={"vertices_count", "texture_coordinates_count"},
):
    """
    a 3D mesh consisting of vertices and faces
    with optional colors and texture coordinates
    """

    vertices: List[float]
    faces: List[int]
    colors: List[int] = field(default_factory=list)
    textureCoordinates: List[float] = field(default_factory=list)
    vertexNormals: List[float] = field(default_factory=list)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"vertices: {self.vertices_count}, "
            f"units: {self.units}, "
            f"has_colors: {len(self.colors) > 0}, "
            f"has_texture_coords: {len(self.textureCoordinates) > 0})"
        )

    @property
    def vertices_count(self) -> int:
        """
        get the number of vertices in the mesh
        """

        if len(self.vertices) % 3 != 0:
            raise ValueError(
                f"Invalid vertices list: length {len(self.vertices)} "
                f"must be a multiple of 3"
            )
        return len(self.vertices) // 3

    @property
    def texture_coordinates_count(self) -> int:
        """
        get the number of texture coordinates in the mesh
        """
        return len(self.textureCoordinates) // 2

    @property
    def area(self) -> float:
        return self.__dict__.get("_area", 0.0)

    @area.setter
    def area(self, value: float) -> None:
        self.__dict__["_area"] = value

    @property
    def volume(self) -> float:
        return self.__dict__.get("_volume", 0.0)

    @volume.setter
    def volume(self, value: float) -> None:
        self.__dict__["_volume"] = value

    def calculate_area(self) -> float:
        """
        calculate total surface area of the mesh
        """
        total_area = 0.0
        face_index = 0
        i = 0

        while i < len(self.faces):
            vertex_count = self.faces[i]
            if vertex_count >= 3:
                face_vertices = self.get_face_vertices(face_index)
                for j in range(1, vertex_count - 1):
                    v0 = face_vertices[0]
                    v1 = face_vertices[j]
                    v2 = face_vertices[j + 1]
                    a = [v1.x - v0.x, v1.y - v0.y, v1.z - v0.z]
                    b = [v2.x - v0.x, v2.y - v0.y, v2.z - v0.z]
                    cx = a[1] * b[2] - a[2] * b[1]
                    cy = a[2] * b[0] - a[0] * b[2]
                    cz = a[0] * b[1] - a[1] * b[0]
                    area = 0.5 * (cx * cx + cy * cy + cz * cz) ** 0.5
                    total_area += area
            i += vertex_count + 1
            face_index += 1

        return total_area

    def calculate_volume(self) -> float:
        """
        calculate volume of the mesh if it is closed
        """
        if not self.is_closed():
            return 0.0

        total_volume = 0.0
        face_index = 0
        i = 0
        while i < len(self.faces):
            vertex_count = self.faces[i]
            if vertex_count >= 3:
                face_vertices = self.get_face_vertices(face_index)
                v0 = face_vertices[0]
                for j in range(1, vertex_count - 1):
                    v1 = face_vertices[j]
                    v2 = face_vertices[j + 1]
                    a = [v0.x, v0.y, v0.z]
                    b = [v1.x - v0.x, v1.y - v0.y, v1.z - v0.z]
                    c = [v2.x - v0.x, v2.y - v0.y, v2.z - v0.z]
                    cx = b[1] * c[2] - b[2] * c[1]
                    cy = b[2] * c[0] - b[0] * c[2]
                    cz = b[0] * c[1] - b[1] * c[0]
                    v = (a[0] * cx + a[1] * cy + a[2] * cz) / 6.0
                    total_volume += v
            i += vertex_count + 1
            face_index += 1

        return abs(total_volume)

    def get_point(self, index: int) -> Point:
        """
        get vertex at index as a Point object
        """
        if index < 0 or index >= self.vertices_count:
            raise IndexError(f"Vertex index {index} out of range")

        index *= 3
        return Point(
            x=self.vertices[index],
            y=self.vertices[index + 1],
            z=self.vertices[index + 2],
            units=self.units,
        )

    def get_points(self) -> List[Point]:
        """
        get all vertices as Point objects
        """
        return [self.get_point(i) for i in range(self.vertices_count)]

    def get_texture_coordinate(self, index: int) -> Tuple[float, float]:
        """
        get texture coordinate at index
        """
        if index < 0 or index >= self.texture_coordinates_count:
            raise IndexError(f"Texture coordinate index {index} out of range")

        index *= 2
        return (self.textureCoordinates[index], self.textureCoordinates[index + 1])

    def get_face_vertices(self, face_index: int) -> List[Point]:
        """
        get the vertices of a specific face
        """
        i = 0
        current_face = 0

        while i < len(self.faces):
            if current_face == face_index:
                vertex_count = self.faces[i]
                vertices = []
                for j in range(vertex_count):
                    vertex_index = self.faces[i + j + 1]
                    if vertex_index >= self.vertices_count:
                        raise IndexError(f"Vertex index {vertex_index} out of range")
                    vertices.append(self.get_point(vertex_index))
                return vertices

            vertex_count = self.faces[i]
            i += vertex_count + 1
            current_face += 1

        raise IndexError(f"Face index {face_index} out of range")

    def is_closed(self) -> bool:
        """
        check if the mesh is closed (verifying each edge appears twice)
        """
        edge_counts = {}

        i = 0
        while i < len(self.faces):
            vertex_count = self.faces[i]
            for j in range(vertex_count):
                v1 = self.faces[i + 1 + j]
                v2 = self.faces[i + 1 + ((j + 1) % vertex_count)]
                edge = tuple(sorted([v1, v2]))
                edge_counts[edge] = edge_counts.get(edge, 0) + 1

            i += vertex_count + 1

        return all(count == 2 for count in edge_counts.values())



from dataclasses import dataclass

from specklepy.objects.base import Base


@dataclass(kw_only=True)
class Interval(
    Base, speckle_type="Objects.Primitive.Interval", serialize_ignore={"length"}
):
    start: float = 0.0
    end: float = 0.0

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(start: {self.start}, end: {self.end})"

    @property
    def length(self) -> float:
        return abs(self.end - self.start)

    @classmethod
    def unit_interval(cls) -> "Interval":
        interval = cls(start=0, end=1)
        return interval



from dataclasses import dataclass

from specklepy.objects.base import Base
from specklepy.objects.interfaces import IHasUnits


@dataclass(kw_only=True)
class Point(Base, IHasUnits, speckle_type="Objects.Geometry.Point"):
    """
    a 3-dimensional point
    """

    x: float
    y: float
    z: float

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"x: {self.x}, "
            f"y: {self.y}, "
            f"z: {self.z}, "
            f"units: {self.units})"
        )

    def distance_to(self, other: "Point") -> float:
        """
        calculates the distance between this point and another given point.
        """
        if not isinstance(other, Point):
            raise TypeError(f"Expected Point object, got {type(other)}")

        # we assume that host application units are the same for both points
        # unit conversion could be expensive, so we avoid it here
        dx = other.x - self.x
        dy = other.y - self.y
        dz = other.z - self.z

        return (dx * dx + dy * dy + dz * dz) ** 0.5



import math
from dataclasses import dataclass

from specklepy.objects.base import Base
from specklepy.objects.geometry.plane import Plane
from specklepy.objects.geometry.point import Point
from specklepy.objects.interfaces import ICurve, IHasArea, IHasUnits


@dataclass(kw_only=True)
class Circle(Base, IHasUnits, ICurve, IHasArea, speckle_type="Objects.Geometry.Circle"):
    """
    a circular curve based on a plane
    """

    plane: Plane
    center: Point
    radius: float

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"plane: {self.plane}, "
            f"center: {self.center}, "
            f"radius: {self.radius}, "
            f"units: {self.units})"
        )

    @property
    def length(self) -> float:
        return 2 * math.pi * self.radius

    @property
    def area(self) -> float:
        return math.pi * self.radius**2



from dataclasses import dataclass

from specklepy.objects.base import Base


@dataclass(kw_only=True)
class RenderMaterial(
    Base,
    speckle_type="Objects.Other.RenderMaterial",
):
    """
    Minimal physically based material DTO class. Based on references from
    https://threejs.org/docs/index.html#api/en/materials/MeshStandardMaterial
    """

    name: str
    opacity: float = 1.0
    metalness: float = 0.0
    roughness: float = 1.0
    diffuse: int  # ARGB color as int
    emissive: int = 0  # ARGB color as int, defaults to black



import json

import pytest

from specklepy.api import operations
from specklepy.api.client import SpeckleClient
from specklepy.core.api.models.current import Project
from specklepy.objects.base import Base
from specklepy.objects.geometry import Point
from specklepy.transports.memory import MemoryTransport
from specklepy.transports.server import ServerTransport

from .fakemesh import FakeMesh


@pytest.mark.run(order=5)
class TestSerialization:
    def test_serialize(self, base: Base):
        serialized = operations.serialize(base)
        deserialized = operations.deserialize(serialized)

        assert base.get_id() == deserialized.get_id()
        assert base.units == "millimetres"
        assert isinstance(base.test_bases[0], Base)
        assert base["@revit_thing"].speckle_type == "SpecialRevitFamily"
        assert base["@detach"].applicationId == deserialized["@detach"].applicationId

    def test_detaching(self, mesh: FakeMesh):
        transport = MemoryTransport()
        serialized = operations.serialize(mesh, [transport])
        deserialized = operations.deserialize(serialized, transport)

        serialized_dict = json.loads(serialized)

        assert serialized_dict["detach_this"]["speckle_type"] == "reference"
        assert serialized_dict["@detach"]["speckle_type"] == "reference"
        assert serialized_dict["origin"]["speckle_type"] == "reference"
        assert serialized_dict["@detached_list"][-1]["speckle_type"] == "reference"
        assert mesh.get_id() == deserialized.get_id()

    def test_chunking(self, mesh: FakeMesh):
        transport = MemoryTransport()
        serialized = operations.serialize(mesh, [transport])
        deserialized = operations.deserialize(serialized, transport)

        serialized_dict = json.loads(serialized)

        assert len(serialized_dict["vertices"]) == 3
        assert len(serialized_dict["@(100)colours"]) == 3
        assert len(serialized_dict["@()default_chunk"]) == 1
        assert serialized_dict["vertices"][0]["speckle_type"] == "reference"
        assert serialized_dict["@(100)colours"][0]["speckle_type"] == "reference"
        assert serialized_dict["@()default_chunk"][0]["speckle_type"] == "reference"
        assert mesh.get_id() == deserialized.get_id()

    def test_send_and_receive(
        self, client: SpeckleClient, sample_project: Project, mesh: FakeMesh
    ):
        transport = ServerTransport(client=client, stream_id=sample_project.id)
        hash = operations.send(mesh, transports=[transport])

        # also try constructing server transport with token and url
        transport = ServerTransport(
            stream_id=sample_project.id, token=client.account.token, url=client.url
        )
        # use a fresh memory transport to force receiving from remote
        received = operations.receive(
            hash, remote_transport=transport, local_transport=MemoryTransport()
        )

        assert isinstance(received, FakeMesh)
        assert received.vertices == mesh.vertices
        assert isinstance(received.origin, Point)
        assert received.origin.x == mesh.origin.x
        # not comparing hashes as order is not guaranteed back from server

        mesh.id = hash  # populate with decomposed id for use in proceeding tests

    def test_receive_local(self, client: SpeckleClient, mesh: FakeMesh):
        hash = operations.send(mesh)  # defaults to SQLiteTransport
        received = operations.receive(hash)

        assert isinstance(received, Base)
        assert mesh.get_id() == received.get_id()

    def test_unknown_type(self):
        unknown = '{"speckle_type": "mysterious.type"}'
        deserialised = operations.deserialize(unknown)

        assert isinstance(deserialised, Base)
        assert deserialised.speckle_type == "mysterious.type"

    def test_no_speckle_type(self):
        untyped = '{"foo": "bar"}'
        deserialised = operations.deserialize(untyped)

        assert deserialised == {"foo": "bar"}

    def test_big_int(self):
        big_int = '{"big": ' + str(2**64) + "}"
        deserialised = operations.deserialize(big_int)

        assert deserialised == {"big": 2**64}



import pytest

from specklepy.core.api.operations import deserialize, serialize
from specklepy.logging.exceptions import SpeckleException
from specklepy.objects.base import Base
from specklepy.objects.data_objects import BlenderObject, DataObject, QgisObject
from specklepy.objects.interfaces import (
    IBlenderObject,
    IDataObject,
    IGisObject,
    IHasUnits,
)
from specklepy.objects.models.units import Units


def test_data_object_creation():
    display_value = [Base()]
    data_obj = DataObject(
        name="Test Data Object",
        properties={"key1": "value1", "key2": 2},
        displayValue=display_value,
    )

    assert data_obj.name == "Test Data Object"
    assert data_obj.properties == {"key1": "value1", "key2": 2}
    assert data_obj.displayValue == display_value
    assert data_obj.speckle_type == "Objects.Data.DataObject"


def test_inheritance_relationships():
    data_obj = DataObject(
        name="Test Data Object",
        properties={"key": "value"},
        displayValue=[Base()],
    )
    assert isinstance(data_obj, DataObject)
    assert isinstance(data_obj, Base)
    assert isinstance(data_obj, IDataObject)

    qgis_obj = QgisObject(
        name="Test QGIS Object",
        properties={"key": "value"},
        displayValue=[Base()],
        type="Feature",
        units=Units.m,
    )
    assert isinstance(qgis_obj, QgisObject)
    assert isinstance(qgis_obj, DataObject)
    assert isinstance(qgis_obj, Base)
    assert isinstance(qgis_obj, IDataObject)
    assert isinstance(qgis_obj, IGisObject)
    assert isinstance(qgis_obj, IHasUnits)

    blender_obj = BlenderObject(
        name="Test Blender Object",
        properties={"key": "value"},
        displayValue=[Base()],
        type="Mesh",
        units=Units.m,
    )
    assert isinstance(blender_obj, BlenderObject)
    assert isinstance(blender_obj, DataObject)
    assert isinstance(blender_obj, Base)
    assert isinstance(blender_obj, IDataObject)
    assert isinstance(blender_obj, IBlenderObject)
    assert isinstance(blender_obj, IHasUnits)


def test_data_object_invalid_types():
    data_obj = DataObject(
        name="Test Object",
        properties={"key": "value"},
        displayValue=[Base()],
    )

    class ComplexObject:
        def __str__(self):
            raise ValueError("Can't convert to string")

    complex_obj = ComplexObject()

    with pytest.raises((ValueError, SpeckleException)):
        data_obj.name = complex_obj  # should be string

    with pytest.raises(SpeckleException):
        data_obj.properties = [1, 2, 3]  # should be dict, not list

    with pytest.raises(SpeckleException):
        data_obj.displayValue = {"key": "value"}  # should be list, not dict


def test_data_object_serialization():
    display_value = [Base()]
    data_obj = DataObject(
        name="Test Data Object",
        properties={"key1": "value1", "key2": 2},
        displayValue=display_value,
    )

    serialized = serialize(data_obj)
    deserialized = deserialize(serialized)

    assert isinstance(deserialized, DataObject)
    assert deserialized.name == data_obj.name
    assert deserialized.properties == data_obj.properties
    assert len(deserialized.displayValue) == len(data_obj.displayValue)
    assert deserialized.speckle_type == data_obj.speckle_type


def test_qgis_object_creation():
    display_value = [Base()]
    qgis_obj = QgisObject(
        name="Test QGIS Object",
        properties={"key1": "value1"},
        displayValue=display_value,
        type="Feature",
        units=Units.m,
    )

    assert qgis_obj.name == "Test QGIS Object"
    assert qgis_obj.properties == {"key1": "value1"}
    assert qgis_obj.displayValue == display_value
    assert qgis_obj.type == "Feature"
    assert qgis_obj.units == Units.m.value
    assert "Objects.Data.QgisObject" in qgis_obj.speckle_type


def test_qgis_object_serialization():
    display_value = [Base()]
    qgis_obj = QgisObject(
        name="Test QGIS Object",
        properties={"key1": "value1"},
        displayValue=display_value,
        type="Feature",
        units=Units.m,
    )

    serialized = serialize(qgis_obj)
    deserialized = deserialize(serialized)

    assert isinstance(deserialized, QgisObject)
    assert deserialized.name == qgis_obj.name
    assert deserialized.properties == qgis_obj.properties
    assert len(deserialized.displayValue) == len(qgis_obj.displayValue)
    assert deserialized.type == qgis_obj.type
    assert deserialized.units == qgis_obj.units
    assert "Objects.Data.QgisObject" in deserialized.speckle_type


def test_blender_object_creation():
    display_value = [Base()]
    blender_obj = BlenderObject(
        name="Test Blender Object",
        properties={"key1": "value1"},
        displayValue=display_value,
        type="Mesh",
        units=Units.m,
    )

    assert blender_obj.name == "Test Blender Object"
    assert blender_obj.properties == {"key1": "value1"}
    assert blender_obj.displayValue == display_value
    assert blender_obj.type == "Mesh"
    assert blender_obj.units == Units.m.value
    assert "Objects.Data.BlenderObject" in blender_obj.speckle_type


def test_blender_object_invalid_types():
    blender_obj = BlenderObject(
        name="Test Object",
        properties={"key": "value"},
        displayValue=[Base()],
        type="Mesh",
        units=Units.m,
    )

    class ComplexObject:
        def __str__(self):
            raise ValueError("Can't convert to string")

    complex_obj = ComplexObject()

    with pytest.raises((ValueError, SpeckleException)):
        blender_obj.type = complex_obj  # should be string


def test_blender_object_serialization():
    display_value = [Base()]
    blender_obj = BlenderObject(
        name="Test Blender Object",
        properties={"key1": "value1"},
        displayValue=display_value,
        type="Mesh",
        units=Units.m,
    )

    serialized = serialize(blender_obj)
    deserialized = deserialize(serialized)

    assert isinstance(deserialized, BlenderObject)
    assert deserialized.name == blender_obj.name
    assert deserialized.properties == blender_obj.properties
    assert len(deserialized.displayValue) == len(blender_obj.displayValue)
    assert deserialized.type == blender_obj.type
    assert deserialized.units == blender_obj.units
    assert "Objects.Data.BlenderObject" in deserialized.speckle_type


def test_data_object_property_modification():
    data_obj = DataObject(
        name="Original Name",
        properties={"original": "value"},
        displayValue=[Base()],
    )

    data_obj.name = "Updated Name"
    data_obj.properties = {"updated": "property"}
    new_display_value = [Base(), Base()]
    data_obj.displayValue = new_display_value

    assert data_obj.name == "Updated Name"
    assert data_obj.properties == {"updated": "property"}
    assert data_obj.displayValue == new_display_value


def test_qgis_object_property_modification():
    """Test modification of QgisObject properties after creation."""
    qgis_obj = QgisObject(
        name="Original Name",
        properties={"original": "value"},
        displayValue=[Base()],
        type="OriginalType",
        units=Units.m,
    )

    qgis_obj.type = "UpdatedType"

    assert qgis_obj.type == "UpdatedType"


def test_blender_object_property_modification():
    blender_obj = BlenderObject(
        name="Original Name",
        properties={"original": "value"},
        displayValue=[Base()],
        type="OriginalType",
        units=Units.m,
    )

    blender_obj.type = "UpdatedType"

    assert blender_obj.type == "UpdatedType"



from pydantic import AliasGenerator, BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class GraphQLBaseModel(BaseModel):
    """
    Parent class for all GraphQL Object Model classes
    Sets-up a pydantic config to serialize properties using a camel case alias
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(
            serialization_alias=to_camel,
            validation_alias=to_camel,
        ),
        populate_by_name=True,
    )



"""This module provides an abstraction layer above the Speckle Automate runtime."""

import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import httpx
from gql import gql

from speckle_automate.schema import (
    AutomateBase,
    AutomationResult,
    AutomationRunData,
    AutomationStatus,
    ObjectResultLevel,
    ResultCase,
)
from specklepy.api import operations
from specklepy.api.client import SpeckleClient
from specklepy.core.api.inputs.model_inputs import CreateModelInput
from specklepy.core.api.inputs.version_inputs import CreateVersionInput
from specklepy.core.api.models.current import Model, Version
from specklepy.logging.exceptions import SpeckleException
from specklepy.objects.base import Base
from specklepy.transports.memory import MemoryTransport
from specklepy.transports.server import ServerTransport


@dataclass
class AutomationContext:
    """A context helper class.

    This class exposes methods to work with the Speckle Automate context inside
    Speckle Automate functions.

    An instance of AutomationContext is injected into every run of a function.
    """

    automation_run_data: AutomationRunData
    speckle_client: SpeckleClient
    _server_transport: ServerTransport
    _speckle_token: str

    #: keep a memory transponrt at hand, to speed up things if needed
    _memory_transport: MemoryTransport = field(default_factory=MemoryTransport)

    #: added for performance measuring
    _init_time: float = field(default_factory=time.perf_counter)
    _automation_result: AutomationResult = field(default_factory=AutomationResult)

    @classmethod
    def initialize(
        cls, automation_run_data: Union[str, AutomationRunData], speckle_token: str
    ) -> "AutomationContext":
        """Bootstrap the AutomateSDK from raw data.

        Todo:
        ----
            * bootstrap a structlog logger instance
            * expose a logger, that ppl can use instead of print
        """
        # parse the json value if its not an initialized project data instance
        automation_run_data = (
            automation_run_data
            if isinstance(automation_run_data, AutomationRunData)
            else AutomationRunData.model_validate_json(automation_run_data)
        )
        speckle_client = SpeckleClient(
            automation_run_data.speckle_server_url,
            automation_run_data.speckle_server_url.startswith("https"),
        )
        speckle_client.authenticate_with_token(speckle_token)
        if not speckle_client.account:
            msg = (
                f"Could not authenticate to {automation_run_data.speckle_server_url}",
                "with the provided token",
            )
            raise ValueError(msg)
        server_transport = ServerTransport(
            automation_run_data.project_id, speckle_client
        )
        return cls(automation_run_data, speckle_client, server_transport, speckle_token)

    @property
    def run_status(self) -> AutomationStatus:
        """Get the status of the automation run."""
        return self._automation_result.run_status

    @property
    def status_message(self) -> Optional[str]:
        """Get the current status message."""
        return self._automation_result.status_message

    def elapsed(self) -> float:
        """Return the elapsed time in seconds since the initialization time."""
        return time.perf_counter() - self._init_time

    def receive_version(self) -> Base:
        """Receive the Speckle project version that triggered this automation run."""
        # TODO: this is a quick hack to keep implementation consistency.
        # Move to proper receive many versions
        version_id = self.automation_run_data.triggers[0].payload.version_id
        try:
            version = self.speckle_client.version.get(
                version_id, self.automation_run_data.project_id
            )
        except SpeckleException as err:
            raise ValueError(
                f"""Could not receive specified version.
                Is your environment configured correctly?
                project_id: {self.automation_run_data.project_id}
                model_id: {self.automation_run_data.triggers[0].payload.model_id}
                version_id: {self.automation_run_data.triggers[0].payload.version_id}
                """
            ) from err

        if not version.referenced_object:
            raise Exception(
                "This version is past the version history limit,",
                " cannot execute an automation on it",
            )

        base = operations.receive(
            version.referenced_object, self._server_transport, self._memory_transport
        )
        # self._closure_tree = base["__closure"]
        print(
            f"It took {self.elapsed():.2f} seconds to receive",
            f" the speckle version {version_id}",
        )
        return base

    def create_new_model_in_project(
        self, model_name: str, model_description: Optional[str] = None
    ) -> Model:
        input = CreateModelInput(
            name=model_name,
            description=model_description,
            project_id=self.automation_run_data.project_id,
        )

        return self.speckle_client.model.create(input)

    def get_model(self, model_id: str) -> Model:
        """
        Args:
            model_id (str): The id of the model to get
        """
        return self.speckle_client.model.get(
            model_id, self.automation_run_data.project_id
        )

    def create_new_version_in_project(
        self, root_object: Base, model_id: str, version_message: str = ""
    ) -> Version:
        """Save a base model to a new version on the project.

        Args:
            root_object (Base): The Speckle base object for the new version.
            model_id (str): Id of model to create the new version on.
            version_message (str): The message for the new version.
        """

        matching_trigger = [
            t
            for t in self.automation_run_data.triggers
            if t.payload.model_id == model_id
        ]
        if matching_trigger:
            raise ValueError(
                f"The target model: {model_id} cannot match the model"
                f" that triggered this automation:"
                f" {matching_trigger[0].payload.model_id}"
            )

        root_object_id = operations.send(
            root_object,
            [self._server_transport, self._memory_transport],
            use_default_cache=False,
        )

        create_version_input = CreateVersionInput(
            object_id=root_object_id,
            model_id=model_id,
            project_id=self.automation_run_data.project_id,
            message=version_message,
            source_application="SpeckleAutomate",
        )
        version = self.speckle_client.version.create(create_version_input)

        self._automation_result.result_versions.append(version.id)
        return version

    @property
    def context_view(self) -> Optional[str]:
        return self._automation_result.result_view

    def set_context_view(
        self,
        # f"{model_id}@{version_id} or {model_id} "
        resource_ids: Optional[List[str]] = None,
        include_source_model_version: bool = True,
    ) -> None:
        link_resources = (
            [
                f"{t.payload.model_id}@{t.payload.version_id}"
                for t in self.automation_run_data.triggers
            ]
            if include_source_model_version
            else []
        )
        if resource_ids:
            link_resources.extend(resource_ids)
        if not link_resources:
            raise Exception(
                "We do not have enough resource ids to compose a context view"
            )
        self._automation_result.result_view = (
            f"/projects/{self.automation_run_data.project_id}"
            f"/models/{','.join(link_resources)}"
        )

    def report_run_status(self) -> None:
        """Report the current run status to the project of this automation."""
        query = gql(
            """
            mutation AutomateFunctionRunStatusReport(
                $projectId: String!
                $functionRunId: String!
                $status: AutomateRunStatus!
                $statusMessage: String
                $results: JSONObject
                $contextView: String
            ){
                automateFunctionRunStatusReport(input: {
                    projectId: $projectId
                    functionRunId: $functionRunId
                    status: $status
                    statusMessage: $statusMessage
                    contextView: $contextView
                    results: $results
                })
            }
        """
        )
        if self.run_status in [AutomationStatus.SUCCEEDED, AutomationStatus.FAILED]:
            object_results = {
                "version": 2,
                "values": {
                    "objectResults": self._automation_result.model_dump(by_alias=True)[
                        "objectResults"
                    ],
                    "blobIds": self._automation_result.blobs,
                },
            }
        else:
            object_results = None

        params = {
            "projectId": self.automation_run_data.project_id,
            "functionRunId": self.automation_run_data.function_run_id,
            "status": self.run_status.value,
            "statusMessage": self._automation_result.status_message,
            "results": object_results,
            "contextView": self._automation_result.result_view,
        }
        print(f"Reporting run status with content: {params}")
        self.speckle_client.httpclient.execute(query, params)

    def store_file_result(self, file_path: Union[Path, str]) -> None:
        """Save a file attached to the project of this automation."""
        path_obj = (
            Path(file_path).resolve() if isinstance(file_path, str) else file_path
        )

        class UploadResult(AutomateBase):
            blob_id: str
            file_name: str
            upload_status: int

        class BlobUploadResponse(AutomateBase):
            upload_results: list[UploadResult]

        if not path_obj.exists():
            raise ValueError("The given file path doesn't exist")

        files = {path_obj.name: path_obj.open("rb")}

        url = (
            f"{self.automation_run_data.speckle_server_url}api/stream/"
            f"{self.automation_run_data.project_id}/blob"
        )
        data = (
            httpx.post(
                url,
                files=files,
                headers={"authorization": f"Bearer {self._speckle_token}"},
            )
            .raise_for_status()
            .json()
        )

        upload_response = BlobUploadResponse.model_validate(data)

        if len(upload_response.upload_results) != 1:
            raise ValueError("Expecting one upload result.")

        self._automation_result.blobs.extend(
            [upload_result.blob_id for upload_result in upload_response.upload_results]
        )

    def mark_run_failed(self, status_message: str) -> None:
        """Mark the current run a failure."""
        self._mark_run(AutomationStatus.FAILED, status_message)

    def mark_run_exception(self, status_message: str) -> None:
        """Mark the current run a failure."""
        self._mark_run(AutomationStatus.EXCEPTION, status_message)

    def mark_run_success(self, status_message: Optional[str]) -> None:
        """Mark the current run a success with an optional message."""
        self._mark_run(AutomationStatus.SUCCEEDED, status_message)

    def _mark_run(
        self, status: AutomationStatus, status_message: Optional[str]
    ) -> None:
        duration = self.elapsed()
        self._automation_result.status_message = status_message
        self._automation_result.run_status = status
        self._automation_result.elapsed = duration

        msg = f"Automation run {status.value} after {duration:.2f} seconds."
        print("\n".join([msg, status_message]) if status_message else msg)

    def attach_error_to_objects(
        self,
        category: str,
        affected_objects: Union[Base, List[Base]],
        message: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        visual_overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a new error case to the run results.
        Args:
            category (str): A short tag for the event type.
            affected_objects (Union[Base, List[Base]]): A single object or a list of
                objects that are causing the error case.
            message (Optional[str]): Optional message.
            metadata: User provided metadata key value pairs
            visual_overrides: Case specific 3D visual overrides.
        """
        self.attach_result_to_objects(
            ObjectResultLevel.ERROR,
            category,
            affected_objects,
            message,
            metadata,
            visual_overrides,
        )

    def attach_warning_to_objects(
        self,
        category: str,
        affected_objects: Union[Base, List[Base]],
        message: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        visual_overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a new warning case to the run results.

        Args:
            category (str): A short tag for the event type.
            affected_objects (Union[Base, List[Base]]): A single object or a list of
                objects that are causing the warning case.
            message (Optional[str]): Optional message.
            metadata: User provided metadata key value pairs
            visual_overrides: Case specific 3D visual overrides.
        """
        self.attach_result_to_objects(
            ObjectResultLevel.WARNING,
            category,
            affected_objects,
            message,
            metadata,
            visual_overrides,
        )

    def attach_success_to_objects(
        self,
        category: str,
        affected_objects: Union[Base, List[Base]],
        message: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        visual_overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a new success case to the run results.

        Args:
            category (str): A short tag for the event type.
            affected_objects (Union[Base, List[Base]]): A single object or a list of
                objects that are causing the success case.
            message (Optional[str]): Optional message.
            metadata: User provided metadata key value pairs
            visual_overrides: Case specific 3D visual overrides.
        """
        self.attach_result_to_objects(
            ObjectResultLevel.SUCCESS,
            category,
            affected_objects,
            message,
            metadata,
            visual_overrides,
        )

    def attach_info_to_objects(
        self,
        category: str,
        affected_objects: Union[Base, List[Base]],
        message: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        visual_overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a new info case to the run results.

        Args:
            category (str): A short tag for the event type.
            affected_objects (Union[Base, List[Base]]): A single object or a list of
                objects that are causing the info case.
            message (Optional[str]): Optional message.
            metadata: User provided metadata key value pairs
            visual_overrides: Case specific 3D visual overrides.
        """
        self.attach_result_to_objects(
            ObjectResultLevel.INFO,
            category,
            affected_objects,
            message,
            metadata,
            visual_overrides,
        )

    def attach_result_to_objects(
        self,
        level: ObjectResultLevel,
        category: str,
        affected_objects: Union[Base, List[Base]],
        message: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        visual_overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a new result case to the run results.

        Args:
            level: Result level.
            category (str): A short tag for the event type.
            affected_objects (Union[Base, List[Base]]): A single object or a list of
                objects that are causing the info case.
            message (Optional[str]): Optional message.
            metadata: User provided metadata key value pairs
            visual_overrides: Case specific 3D visual overrides.
        """
        if isinstance(affected_objects, list):
            if len(affected_objects) < 1:
                raise ValueError(
                    f"Need atleast one object to report a(n) {level.value.upper()}"
                )
            object_list = affected_objects
        else:
            object_list = [affected_objects]

        ids: Dict[str, Optional[str]] = {}
        for o in object_list:
            # validate that the Base.id is not None. If its a None, throw an Exception
            if not o.id:
                raise Exception(
                    f"You can only attach {level} results to objects with an id."
                )
            ids[o.id] = o.applicationId
        print(
            f"Created new {level.value.upper()}"
            f" category: {category} caused by: {message}"
        )
        self._automation_result.object_results.append(
            ResultCase(
                category=category,
                level=level,
                object_app_ids=ids,
                message=message,
                metadata=metadata,
                visual_overrides=visual_overrides,
            )
        )



"""This module contains an SDK for working with Speckle Automate."""

from speckle_automate.automation_context import AutomationContext
from speckle_automate.runner import execute_automate_function, run_function
from speckle_automate.schema import (
    AutomateBase,
    AutomationResult,
    AutomationRunData,
    AutomationStatus,
    ObjectResultLevel,
    ResultCase,
)

__all__ = [
    "AutomationContext",
    "AutomateBase",
    "AutomationStatus",
    "AutomationResult",
    "AutomationRunData",
    "ResultCase",
    "ObjectResultLevel",
    "run_function",
    "execute_automate_function",
]



import pytest

from specklepy.objects.models.units import Units, get_scale_factor


@pytest.mark.parametrize(
    "fromUnits, toUnits, inValue, expectedOutValue",
    [
        # To self
        (Units.km, Units.km, 1.5, 1.5),
        (Units.km, Units.km, 0, 0),
        (Units.m, Units.m, 1.5, 1.5),
        (Units.m, Units.m, 0, 0),
        (Units.cm, Units.cm, 1.5, 1.5),
        (Units.cm, Units.cm, 0, 0),
        (Units.mm, Units.mm, 1.5, 1.5),
        (Units.mm, Units.mm, 0, 0),
        (Units.miles, Units.miles, 1.5, 1.5),
        (Units.miles, Units.miles, 0, 0),
        (Units.yards, Units.yards, 1.5, 1.5),
        (Units.yards, Units.yards, 0, 0),
        (Units.feet, Units.feet, 1.5, 1.5),
        (Units.feet, Units.feet, 0, 0),
        # To Meters
        (Units.km, Units.m, 987654.321, 987654321),
        (Units.m, Units.m, 987654.321, 987654.321),
        (Units.mm, Units.m, 98765432.1, 98765.4321),
        (Units.cm, Units.m, 9876543.21, 98765.4321),
        # To negative meters
        (Units.km, Units.m, -987654.321, -987654321),
        (Units.m, Units.m, -987654.321, -987654.321),
        (Units.mm, Units.m, -98765432.1, -98765.4321),
        (Units.cm, Units.m, -9876543.21, -98765.4321),
        (Units.m, Units.km, 987654.321, 987.654321),
        (Units.m, Units.cm, 987654.321, 98765432.1),
        (Units.m, Units.mm, 987654.321, 987654321),
        # Imperial
        (Units.miles, Units.m, 123.45, 198673.517),
        (Units.miles, Units.inches, 123.45, 7821792),
        (Units.yards, Units.m, 123.45, 112.88268),
        (Units.yards, Units.inches, 123.45, 4444.2),
        (Units.feet, Units.m, 123.45, 37.62756),
        (Units.feet, Units.inches, 123.45, 1481.4),
        (Units.inches, Units.m, 123.45, 3.13563),
    ],
)
def test_get_scale_factor_between_units(
    fromUnits: Units, toUnits: Units, inValue: float, expectedOutValue: float
):
    Tolerance = 1e-10
    actual = inValue * get_scale_factor(fromUnits, toUnits)
    assert actual - expectedOutValue < Tolerance



# ignoring "line too long" check from linter
# to match with SpeckleExceptions
# ruff: noqa: E501

import math

import pytest

from specklepy.core.api.operations import deserialize, serialize
from specklepy.logging.exceptions import SpeckleException
from specklepy.objects.geometry import Circle, Plane, Point, Vector
from specklepy.objects.models.units import Units
from specklepy.objects.primitive import Interval


@pytest.fixture
def sample_plane():
    origin = Point(x=0.0, y=0.0, z=0.0, units=Units.m)
    normal = Vector(x=0.0, y=0.0, z=1.0, units=Units.m)
    xdir = Vector(x=1.0, y=0.0, z=0.0, units=Units.m)
    ydir = Vector(x=0.0, y=1.0, z=0.0, units=Units.m)
    return Plane(origin=origin, normal=normal, xdir=xdir, ydir=ydir, units=Units.m)


@pytest.fixture
def sample_center():
    return Point(x=0.0, y=0.0, z=0.0, units=Units.m)


@pytest.fixture
def sample_circle(sample_plane, sample_center):
    return Circle(plane=sample_plane, center=sample_center, radius=1.0, units=Units.m)


def test_circle_creation(sample_plane, sample_center):
    circle = Circle(plane=sample_plane, center=sample_center, radius=1.0, units=Units.m)
    assert circle.plane == sample_plane
    assert circle.center == sample_center
    assert circle.radius == 1.0
    assert circle.units == Units.m.value


def test_circle_domain(sample_circle):
    assert isinstance(sample_circle.domain, Interval)
    assert sample_circle.domain.start == 0.0
    assert sample_circle.domain.end == 1.0


@pytest.mark.parametrize(
    "radius,expected_length", [(1.0, 2 * math.pi), (2.0, 4 * math.pi), (0.5, math.pi)]
)
def test_circle_length(sample_circle, radius, expected_length):
    sample_circle.radius = radius
    assert sample_circle.length == pytest.approx(expected_length)


@pytest.mark.parametrize(
    "radius,expected_area", [(1.0, math.pi), (2.0, 4 * math.pi), (0.5, math.pi * 0.25)]
)
def test_circle_area(sample_circle, radius, expected_area):
    sample_circle.radius = radius
    assert sample_circle.area == pytest.approx(expected_area)


@pytest.mark.parametrize("new_units", ["mm", "cm", "in"])
def test_circle_units(sample_plane, sample_center, new_units):
    circle = Circle(plane=sample_plane, center=sample_center, radius=1.0, units=Units.m)
    assert circle.units == Units.m.value
    circle.units = new_units
    assert circle.units == new_units


@pytest.mark.parametrize(
    "invalid_param, test_params, error_msg",
    [
        (
            "plane",
            {"plane": "not a plane", "center": None, "radius": 1.0},
            "Cannot set 'Circle.plane':it expects type '<class 'specklepy.objects.geometry.plane.Plane'>',but received type 'str'",
        ),
        (
            "center",
            {"plane": None, "center": "not a point", "radius": 1.0},
            "Cannot set 'Circle.center':it expects type '<class 'specklepy.objects.geometry.point.Point'>',but received type 'str'",
        ),
        (
            "radius",
            {"plane": None, "center": None, "radius": "not a number"},
            "Cannot set 'Circle.radius':it expects type '<class 'float'>',but received type 'str'",
        ),
    ],
)
def test_circle_inval(
    sample_plane, sample_center, invalid_param, test_params, error_msg
):
    if invalid_param != "plane":
        test_params["plane"] = sample_plane
    if invalid_param != "center":
        test_params["center"] = sample_center

    with pytest.raises(SpeckleException) as exc_info:
        Circle(**test_params, units=Units.m)
    assert str(exc_info.value) == f"SpeckleException: {error_msg}"


def test_circle_serialization(sample_circle):
    serialized = serialize(sample_circle)
    deserialized = deserialize(serialized)

    assert isinstance(deserialized, Circle)
    assert deserialized.plane.origin.x == sample_circle.plane.origin.x
    assert deserialized.plane.origin.y == sample_circle.plane.origin.y
    assert deserialized.plane.origin.z == sample_circle.plane.origin.z

    assert deserialized.plane.normal.x == sample_circle.plane.normal.x
    assert deserialized.plane.normal.y == sample_circle.plane.normal.y
    assert deserialized.plane.normal.z == sample_circle.plane.normal.z

    assert deserialized.center.x == sample_circle.center.x
    assert deserialized.center.y == sample_circle.center.y
    assert deserialized.center.z == sample_circle.center.z

    assert deserialized.radius == sample_circle.radius
    assert deserialized.units == sample_circle.units



import pytest

from specklepy.core.api.operations import deserialize, serialize
from specklepy.objects.geometry.mesh import Mesh
from specklepy.objects.geometry.point import Point
from specklepy.objects.models.units import Units


@pytest.fixture
def cube_vertices():
    return [
        -0.5,
        -0.5,
        -0.5,
        0.5,
        -0.5,
        -0.5,
        0.5,
        0.5,
        -0.5,
        -0.5,
        0.5,
        -0.5,
        -0.5,
        -0.5,
        0.5,
        0.5,
        -0.5,
        0.5,
        0.5,
        0.5,
        0.5,
        -0.5,
        0.5,
        0.5,
    ]


@pytest.fixture
def cube_faces():
    return [
        4,
        0,
        3,
        2,
        1,  # bottom (-z)
        4,
        4,
        5,
        6,
        7,  # top (+z)
        4,
        0,
        1,
        5,
        4,  # front (-y)
        4,
        3,
        7,
        6,
        2,  # back (+y)
        4,
        0,
        4,
        7,
        3,  # left (-x)
        4,
        1,
        2,
        6,
        5,  # right (+x)
    ]


@pytest.fixture
def cube_colors():
    return [
        255,
        0,
        0,
        255,  # red
        0,
        255,
        0,
        255,  # green
        0,
        0,
        255,
        255,  # blue
        255,
        255,
        0,
        255,  # yellow
        255,
        0,
        255,
        255,  # magenta
        0,
        255,
        255,
        255,  # cyan
        255,
        255,
        255,
        255,  # white
        0,
        0,
        0,
        255,  # black
    ]


@pytest.fixture
def cube_texture_coords():
    return [
        0.0,
        0.0,
        1.0,
        0.0,
        1.0,
        1.0,
        0.0,
        1.0,
        0.0,
        0.0,
        1.0,
        0.0,
        1.0,
        1.0,
        0.0,
        1.0,
    ]


@pytest.fixture
def sample_mesh(cube_vertices, cube_faces):
    return Mesh(vertices=cube_vertices, faces=cube_faces, units=Units.m)


@pytest.fixture
def full_mesh(cube_vertices, cube_faces, cube_colors, cube_texture_coords):
    return Mesh(
        vertices=cube_vertices,
        faces=cube_faces,
        colors=cube_colors,
        textureCoordinates=cube_texture_coords,
        units=Units.m,
    )


def test_mesh_creation(cube_vertices, cube_faces):
    mesh = Mesh(vertices=cube_vertices, faces=cube_faces, units=Units.m)

    assert mesh.vertices == cube_vertices
    assert mesh.faces == cube_faces
    assert mesh.colors == []
    assert mesh.textureCoordinates == []
    assert mesh.vertexNormals == []
    assert mesh.units == Units.m.value


def test_mesh_vertex_count(sample_mesh):
    assert sample_mesh.vertices_count == 8  # cube has 8 vertices


def test_mesh_texture_coordinates_count(full_mesh):
    assert full_mesh.texture_coordinates_count == 8  # one UV per vertex


def test_mesh_get_point(sample_mesh):
    point = sample_mesh.get_point(0)
    assert isinstance(point, Point)
    assert point.x == -0.5
    assert point.y == -0.5
    assert point.z == -0.5
    assert point.units == sample_mesh.units

    with pytest.raises(IndexError):
        sample_mesh.get_point(8)  # Beyond vertex count


def test_mesh_get_points(sample_mesh):
    points = sample_mesh.get_points()
    assert len(points) == 8
    assert all(isinstance(p, Point) for p in points)
    assert all(p.units == sample_mesh.units for p in points)


def test_mesh_get_texture_coordinate(full_mesh):
    uv = full_mesh.get_texture_coordinate(0)
    assert uv == (0.0, 0.0)

    with pytest.raises(IndexError):
        full_mesh.get_texture_coordinate(8)  # beyond UV count


def test_mesh_get_face_vertices(sample_mesh):
    face_vertices = sample_mesh.get_face_vertices(0)
    assert len(face_vertices) == 4
    assert all(isinstance(v, Point) for v in face_vertices)

    with pytest.raises(IndexError):
        sample_mesh.get_face_vertices(6)  # beyond face count


def test_mesh_is_closed(sample_mesh):
    assert sample_mesh.is_closed()  # cube is a closed mesh


def test_mesh_area(sample_mesh):
    calculated_area = sample_mesh.calculate_area()
    sample_mesh.area = calculated_area
    assert sample_mesh.area == pytest.approx(6.0)


def test_mesh_volume(sample_mesh):
    calculated_volume = sample_mesh.calculate_volume()
    sample_mesh.volume = calculated_volume

    # Verify volume is set correctly
    assert sample_mesh.volume == pytest.approx(1.0)


def test_mesh_invalid_vertices():
    mesh = Mesh(vertices=[0.0, 0.0], faces=[3, 0, 1, 2], units=Units.m)

    with pytest.raises(ValueError):
        _ = mesh.vertices_count


def test_mesh_invalid_faces():
    vertices = [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0]
    with pytest.raises(IndexError):
        # Face references vertex index out of range
        mesh = Mesh(vertices=vertices, faces=[3, 0, 1, 5], units=Units.m)
        mesh.get_face_vertices(0)


def test_mesh_serialization(full_mesh):
    serialized = serialize(full_mesh)
    deserialized = deserialize(serialized)

    assert isinstance(deserialized, Mesh)
    assert deserialized.vertices == full_mesh.vertices
    assert deserialized.faces == full_mesh.faces
    assert deserialized.colors == full_mesh.colors
    assert deserialized.textureCoordinates == full_mesh.textureCoordinates
    assert deserialized.units == full_mesh.units



from dataclasses import dataclass
from typing import Dict, List, Optional
from unittest import TestCase

from specklepy.objects.base import Base
from specklepy.objects.graph_traversal.traversal import GraphTraversal, TraversalRule


@dataclass()
class TraversalMock(Base):
    child: Optional[Base]
    list_children: List[Base]
    dict_children: Dict[str, Base]


class GraphTraversalTests(TestCase):
    def test_traverse_list_members(self):
        traverse_lists_rule = TraversalRule(
            [lambda _: True],
            lambda x: [
                item
                for item in x.get_member_names()
                if isinstance(getattr(x, item, None), list)
            ],
        )

        expected_traverse = Base()
        expected_traverse.id = "List Member"

        expected_ignore = Base()
        expected_ignore.id = "Not List Member"

        test_case = TraversalMock(
            list_children=[expected_traverse],
            dict_children={"myprop": expected_ignore},
            child=expected_ignore,
        )

        ret = [
            context.current
            for context in GraphTraversal([traverse_lists_rule]).traverse(test_case)
        ]

        self.assertCountEqual(ret, [test_case, expected_traverse])
        self.assertNotIn(expected_ignore, ret)
        self.assertEqual(len(ret), 2)

    def test_traverse_dict_members(self):
        traverse_lists_rule = TraversalRule(
            [lambda _: True],
            lambda x: [
                item
                for item in x.get_member_names()
                if isinstance(getattr(x, item, None), dict)
            ],
        )

        expected_traverse = Base()
        expected_traverse.id = "Dict Member"

        expected_ignore = Base()
        expected_ignore.id = "Not Dict Member"

        test_case = TraversalMock(
            list_children=[expected_ignore],
            dict_children={"myprop": expected_traverse},
            child=expected_ignore,
        )

        ret = [
            context.current
            for context in GraphTraversal([traverse_lists_rule]).traverse(test_case)
        ]

        self.assertCountEqual(ret, [test_case, expected_traverse])
        self.assertNotIn(expected_ignore, ret)
        self.assertEqual(len(ret), 2)

    def test_traverse_dynamic(self):
        traverse_lists_rule = TraversalRule(
            [lambda _: True], lambda x: x.get_dynamic_member_names()
        )

        expected_traverse = Base()
        expected_traverse.id = "List Member"

        expected_ignore = Base()
        expected_ignore.id = "Not List Member"

        test_case = TraversalMock(
            child=expected_ignore,
            list_children=[expected_ignore],
            dict_children={"myprop": expected_ignore},
        )
        test_case["dynamicChild"] = expected_traverse
        test_case["dynamicListChild"] = [expected_traverse]

        ret = [
            context.current
            for context in GraphTraversal([traverse_lists_rule]).traverse(test_case)
        ]

        self.assertCountEqual(ret, [test_case, expected_traverse, expected_traverse])
        self.assertNotIn(expected_ignore, ret)
        self.assertEqual(len(ret), 3)



"""
Provides uniform and consistent path helpers for `specklepy`
"""

import os
import sys
from importlib import import_module, invalidate_caches
from pathlib import Path
from typing import Optional

_user_data_env_var = "SPECKLE_USERDATA_PATH"


def _path() -> Optional[Path]:
    """Read the user data path override setting."""
    path_override = os.environ.get(_user_data_env_var)
    if path_override:
        return Path(path_override)
    return None


_application_name = "Speckle"


def override_application_name(application_name: str) -> None:
    """Override the global Speckle application name."""
    global _application_name
    _application_name = application_name


def override_application_data_path(path: Optional[str]) -> None:
    """
    Override the global Speckle application data path.

    If the value of path is `None` the environment variable gets deleted.
    """
    if path:
        os.environ[_user_data_env_var] = path
    else:
        os.environ.pop(_user_data_env_var, None)


def _ensure_folder_exists(base_path: Path, folder_name: str) -> Path:
    path = base_path.joinpath(folder_name)
    path.mkdir(exist_ok=True, parents=True)
    return path


def user_application_data_path() -> Path:
    """Get the platform specific user configuration folder path"""
    path_override = _path()
    if path_override:
        return path_override

    try:
        if sys.platform.startswith("win"):
            app_data_path = os.getenv("APPDATA")
            if not app_data_path:
                raise Exception("Cannot get appdata path from environment.")
            return Path(app_data_path)
        else:
            # try getting the standard XDG_DATA_HOME value
            # as that is used as an override
            app_data_path = os.getenv("XDG_DATA_HOME")
            if app_data_path:
                return Path(app_data_path)
            else:
                return _ensure_folder_exists(Path.home(), ".config")
    except Exception as ex:
        raise Exception("Failed to initialize user application data path.") from ex


def user_speckle_folder_path() -> Path:
    """Get the folder where the user's Speckle data should be stored."""
    return _ensure_folder_exists(user_application_data_path(), _application_name)


def user_speckle_connector_installation_path(host_application: str) -> Path:
    """
    Gets a connector specific installation folder.

    In this folder we can put our connector installation and all python packages.
    """
    return _ensure_folder_exists(
        _ensure_folder_exists(user_speckle_folder_path(), "connector_installations"),
        host_application,
    )


print("Starting module dependency installation")
print(sys.executable)

PYTHON_PATH = sys.executable


def connector_installation_path(host_application: str) -> Path:
    connector_installation_path = user_speckle_connector_installation_path(
        host_application
    )
    connector_installation_path.mkdir(exist_ok=True, parents=True)

    # set user modules path at beginning of paths for earlier hit
    if sys.path[0] != connector_installation_path:
        sys.path.insert(0, str(connector_installation_path))

    print(f"Using connector installation path {connector_installation_path}")
    return connector_installation_path


def is_pip_available() -> bool:
    try:
        import_module("pip")  # noqa F401
        return True
    except ImportError:
        return False


def ensure_pip() -> None:
    print("Installing pip... ")

    from subprocess import run

    completed_process = run([PYTHON_PATH, "-m", "ensurepip"])

    if completed_process.returncode == 0:
        print("Successfully installed pip")
    else:
        raise Exception(
            f"Failed to install pip, got {completed_process.returncode} return code"
        )


def get_requirements_path() -> Path:
    # we assume that a requirements.txt exists next to the __init__.py file
    path = Path(Path(__file__).parent, "requirements.txt")
    assert path.exists()
    return path


def install_requirements(host_application: str) -> None:
    # set up addons/modules under the user
    # script path. Here we'll install the
    # dependencies
    path = connector_installation_path(host_application)
    print(f"Installing Speckle dependencies to {path}")

    from subprocess import run

    completed_process = run(
        [
            PYTHON_PATH,
            "-m",
            "pip",
            "install",
            "-t",
            str(path),
            "-r",
            str(get_requirements_path()),
        ],
        capture_output=True,
        text=True,
    )

    if completed_process.returncode != 0:
        m = (
            "Failed to install dependenices through pip, got ",
            f"{completed_process.returncode} return code",
        )
        print(m)
        raise Exception(m)


def install_dependencies(host_application: str) -> None:
    if not is_pip_available():
        ensure_pip()

    install_requirements(host_application)


def _import_dependencies() -> None:
    import_module("specklepy")
    # the code above doesn't work for now, it fails on importing graphql-core
    # despite that, the connector seams to be working as expected
    # But it would be nice to make this solution work
    # it would ensure that all dependencies are fully loaded
    # requirements = get_requirements_path().read_text()
    # reqs = [
    #     req.split(" ; ")[0].split("==")[0].split("[")[0].replace("-", "_")
    #     for req in requirements.split("\n")
    #     if req and not req.startswith(" ")
    # ]
    # for req in reqs:
    #     print(req)
    #     import_module("specklepy")


def ensure_dependencies(host_application: str) -> None:
    try:
        install_dependencies(host_application)
        invalidate_caches()
        _import_dependencies()
        print("Successfully found dependencies")
    except ImportError as e:
        raise Exception(
            "Cannot automatically ensure Speckle dependencies.",
            f"Please try restarting the host application {host_application}!",
        ) from e



import re
import sys
from pathlib import Path

from specklepy.core.helpers import speckle_path_provider


def test_user_application_data_path():
    user_path = str(speckle_path_provider.user_application_data_path())

    if sys.platform.startswith("win"):
        pattern = r"C:\\Users\\.*\\AppData\\Roaming"
    elif sys.platform.startswith("darwin"):
        pattern = "/Users/.*/.config"
    elif sys.platform.startswith("linux"):
        if user_path.startswith("/root"):
            pattern = "/root/.config"
        else:
            pattern = "/home/.*/.config"
    else:
        raise NotImplementedError("Your OS platform is not supported")

    match = re.search(pattern, user_path)
    assert match


def test_user_application_data_path_override():
    path = "/jiberish"
    speckle_path_provider.override_application_data_path(path)

    user_path = speckle_path_provider.user_application_data_path()
    assert Path(path) == user_path

    speckle_path_provider.override_application_data_path(None)

    user_path = speckle_path_provider.user_application_data_path()
    assert Path(path) != user_path


def test_accounts_folder_name_override():
    old_folder_name = speckle_path_provider._accounts_folder_name
    assert old_folder_name == "Accounts"
    new_folder_name = "foobar"
    speckle_path_provider.override_accounts_folder_name(new_folder_name)
    assert speckle_path_provider._accounts_folder_name == new_folder_name
    speckle_path_provider.override_accounts_folder_name(old_folder_name)


def test_connector_installation_path():
    host_application = "test application"
    connector_path = speckle_path_provider.user_speckle_connector_installation_path(
        host_application
    )
    assert "connector_installations" in str(connector_path)
    assert str(connector_path).endswith(host_application)



import random

import pytest

from specklepy.objects.base import Base


@pytest.fixture(scope="session")
def base() -> Base:
    base = Base()
    base.name = "my_base"
    base.units = "millimetres"
    base.null_val = None
    base.null_dict = {"a null val": None}
    base.tuple = (1, 2, "3")
    base.set = {1, 2, "3"}
    base.vertices = [random.uniform(0, 10) for _ in range(1, 120)]
    base.test_bases = [Base(applicationId=str(i)) for i in range(1, 22)]
    base["@detach"] = Base(applicationId="detached base")
    base["@revit_thing"] = Base.of_type(
        "SpecialRevitFamily", applicationId="secret tho"
    )
    return base



from specklepy.objects.base import Base
from specklepy.objects.graph_traversal.traversal import GraphTraversal, TraversalRule

DISPLAY_VALUE_PROPERTY_ALIASES = {"displayValue", "@displayValue"}
ELEMENTS_PROPERTY_ALIASES = {"elements", "@elements"}


def has_display_value(x: Base):
    return any(hasattr(x, alias) for alias in DISPLAY_VALUE_PROPERTY_ALIASES)


def create_default_traversal_function() -> GraphTraversal:
    """
    Traversal func for traversing the root object of a Speckle Model
    """

    convertible_rule = TraversalRule(
        [lambda b: b.speckle_type != "Base", has_display_value],
        lambda _: ELEMENTS_PROPERTY_ALIASES,
    )

    default_rule = TraversalRule(
        [lambda _: True],
        # NOTE: Unlike the C# implementation, this does not ignore Obsolete members
        lambda o: o.get_member_names(),
        False,
    )

    return GraphTraversal([convertible_rule, default_rule])



import pytest

from specklepy.api.client import SpeckleClient
from specklepy.core.api.enums import ProjectVisibility
from specklepy.core.api.inputs.project_inputs import (
    ProjectCreateInput,
    ProjectUpdateInput,
)
from specklepy.core.api.models import Project
from specklepy.core.api.models.current import ProjectPermissionChecks
from specklepy.logging.exceptions import GraphQLException


@pytest.mark.run()
class TestProjectResource:
    @pytest.fixture()
    def test_project(self, client: SpeckleClient) -> Project:
        project = client.project.create(
            ProjectCreateInput(
                name="test project123",
                description="desc",
                visibility=ProjectVisibility.PRIVATE,
            )
        )
        return project

    @pytest.mark.parametrize(
        "name, description, visibility",
        [
            ("Very private project", "My secret project", ProjectVisibility.PRIVATE),
            ("Very discoverable project", None, ProjectVisibility.UNLISTED),
            ("Very public project", None, ProjectVisibility.PUBLIC),
        ],
    )
    def test_project_create(
        self,
        client: SpeckleClient,
        name: str,
        description: str,
        visibility: ProjectVisibility,
    ):
        input = ProjectCreateInput(
            name=name,
            description=description,
            visibility=visibility,
        )
        result = client.project.create(input)

        assert isinstance(result, Project)
        assert result.id is not None
        assert result.name == name
        assert result.description == (description or "")
        # we've disabled creation of public projects for now, they fall back to unlisted
        if visibility == ProjectVisibility.UNLISTED:
            assert result.visibility == ProjectVisibility.PUBLIC
        else:
            assert result.visibility == visibility

    def test_project_get(self, client: SpeckleClient, test_project: Project):
        result = client.project.get(test_project.id)

        assert isinstance(result, Project)
        assert result.id == test_project.id
        assert result.name == test_project.name
        assert result.description == test_project.description
        assert result.visibility == test_project.visibility
        assert result.created_at == test_project.created_at

    def test_project_get_permissions(
        self, client: SpeckleClient, test_project: Project
    ):
        result = client.project.get_permissions(test_project.id)

        assert isinstance(result, ProjectPermissionChecks)
        assert result.can_create_model.authorized is True
        assert result.can_delete.authorized is True

    def test_project_update(self, client: SpeckleClient, test_project: Project):
        new_name = "MY new name"
        new_description = "MY new desc"
        new_visibility = ProjectVisibility.UNLISTED

        update_data = ProjectUpdateInput(
            id=test_project.id,
            name=new_name,
            description=new_description,
            visibility=new_visibility,
        )

        updated_project = client.project.update(update_data)

        assert isinstance(updated_project, Project)
        assert updated_project.id == test_project.id
        assert updated_project.name == new_name
        assert updated_project.description == new_description
        # we've disabled creation of public projects for now, they fall back to unlisted
        if new_visibility == ProjectVisibility.UNLISTED:
            assert updated_project.visibility == ProjectVisibility.PUBLIC
        else:
            assert updated_project.visibility == new_visibility

    def test_project_delete(self, client: SpeckleClient):
        """Test deleting a project."""
        project_to_delete = client.project.create(
            ProjectCreateInput(name="Delete me", description=None, visibility=None)
        )

        response = client.project.delete(project_to_delete.id)
        assert response is True

        with pytest.raises(GraphQLException):
            client.project.get(project_to_delete.id)



from datetime import datetime
from typing import List

import httpx
from pydantic import AliasGenerator, BaseModel, ConfigDict, HttpUrl
from pydantic.alias_generators import to_pascal


class ConnectorFeedBaseModel(BaseModel):
    """
    Parent class for all Connector Feed Object Model classes
    Sets-up a pydantic config to serialize properties using a pascal case alias
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(
            validation_alias=to_pascal,
        ),
        populate_by_name=True,
    )


class ConnectorVersion(ConnectorFeedBaseModel):
    number: str
    url: HttpUrl
    os: int  # this is an enum, it's properly defined in the old v2 SDK (used by Speckle.Manager.Feed)  # noqa: E501
    architecture: int  # These are enums, they are properly defined in the old v2 SDK (used by Speckle.Manager.Feed) # noqa: E501
    date: datetime
    prerelease: bool


class ConnectorVersions(ConnectorFeedBaseModel):
    versions: List[ConnectorVersion]


def get_latest_version(host_app_slug: str, allow_pre_release: bool) -> ConnectorVersion:
    """
    Fetches the JSON feed for the given connector slug and
    Returns the latest version by date - Note, it does not consider semvers!

    Arguments:
        host_app_slug {str} -- the host app slug to query for
        allow_pre_release {bool} -- if false, only stable releases will be considered
    Raises:
        HTTPStatusError: if http request failed
        ValidationError: response was not valid json
        ValueError: The feed contained no connector versions
    """
    connector_versions = get_connector_versions(host_app_slug).versions
    filtered_versions = [
        v for v in connector_versions if allow_pre_release or not v.prerelease
    ]

    return max(filtered_versions, key=lambda x: x.date)


def get_connector_versions(host_app_slug: str) -> ConnectorVersions:
    """
    Fetches the JSON feed for the given slug (v3 feeds only)
    Raises:
        HTTPStatusError: if http request failed
        ValidationError: response was not valid json
    """
    url = f"https://releases.speckle.dev/manager2/feeds/{host_app_slug.lower()}-v3.json"

    res = httpx.get(url).raise_for_status()

    feed_data = ConnectorVersions.model_validate_json(res.text)

    return feed_data



from typing import Optional

import pytest

from specklepy.api.client import SpeckleClient
from specklepy.core.api.enums import ProjectVisibility
from specklepy.core.api.inputs.project_inputs import (
    ProjectCreateInput,
    ProjectInviteCreateInput,
    ProjectInviteUseInput,
    ProjectUpdateRoleInput,
)
from specklepy.core.api.models import (
    LimitedUser,
    PendingStreamCollaborator,
    Project,
    ProjectWithTeam,
)


@pytest.mark.run()
class TestProjectInviteResource:
    @pytest.fixture
    def project(self, client: SpeckleClient):
        return client.project.create(
            ProjectCreateInput(
                name="test", description=None, visibility=ProjectVisibility.PUBLIC
            )
        )

    @pytest.fixture
    def created_invite(
        self, client: SpeckleClient, second_client: SpeckleClient, project: Project
    ):
        input = ProjectInviteCreateInput(
            email=second_client.account.userInfo.email,
            role=None,
            server_role=None,
            userId=None,
        )
        res = client.project_invite.create(project.id, input)
        invites = second_client.active_user.get_project_invites()
        return next(i for i in invites if i.projectId == res.id)

    def test_project_invite_create_by_email(
        self, client: SpeckleClient, second_client: SpeckleClient, project: Project
    ):
        input = ProjectInviteCreateInput(
            email=second_client.account.userInfo.email,
            role=None,
            server_role=None,
            userId=None,
        )
        res = client.project_invite.create(project.id, input)

        invites = second_client.active_user.get_project_invites()
        invite = next(i for i in invites if i.projectId == res.id)

        assert isinstance(res, ProjectWithTeam)
        assert res.id == project.id
        assert len(res.invited_team) == 1

        assert isinstance(invite.user, LimitedUser)
        assert invite.user.id == second_client.account.userInfo.id
        assert invite.token

    def test_project_invite_create_by_user_id(
        self, client: SpeckleClient, second_client: SpeckleClient, project: Project
    ):
        input = ProjectInviteCreateInput(
            email=None,
            role=None,
            server_role=None,
            userId=second_client.account.userInfo.id,
        )
        res = client.project_invite.create(project.id, input)

        assert isinstance(res, ProjectWithTeam)
        assert res.id == project.id
        assert len(res.invited_team) == 1
        invited_team_member = res.invited_team[0].user
        assert isinstance(invited_team_member, LimitedUser)
        assert invited_team_member.id == second_client.account.userInfo.id

    def test_project_invite_get(
        self,
        second_client: SpeckleClient,
        project: Project,
        created_invite: PendingStreamCollaborator,
    ):
        collaborator = second_client.project_invite.get(
            project.id, created_invite.token
        )
        assert isinstance(collaborator, PendingStreamCollaborator)
        assert collaborator.invite_id == created_invite.invite_id

        assert isinstance(collaborator.user, LimitedUser)
        assert isinstance(created_invite.user, LimitedUser)

        assert collaborator.user.id == created_invite.user.id

    def test_project_invite_get_non_existing(
        self, second_client: SpeckleClient, project: Project
    ):
        collaborator = second_client.project_invite.get(
            project.id, "this is not a real token"
        )

        assert collaborator is None

    def test_project_invite_use_member_added(
        self,
        client: SpeckleClient,
        second_client: SpeckleClient,
        project: Project,
        created_invite: PendingStreamCollaborator,
    ):
        assert created_invite.token

        input = ProjectInviteUseInput(
            accept=True, project_id=created_invite.projectId, token=created_invite.token
        )
        res = second_client.project_invite.use(input)

        assert res is True

        project = client.project.get_with_team(project.id)
        assert isinstance(project, ProjectWithTeam)

        team_members = [c.user.id for c in project.team]
        expected_team_members = [
            client.account.userInfo.id,
            second_client.account.userInfo.id,
        ]

        assert set(team_members) == set(expected_team_members)

    def test_project_invite_cancel_member_not_added(
        self, client: SpeckleClient, created_invite: PendingStreamCollaborator
    ):
        res = client.project_invite.cancel(
            created_invite.projectId, created_invite.invite_id
        )

        assert isinstance(res, ProjectWithTeam)
        assert len(res.invited_team) == 0

    @pytest.mark.parametrize(
        "new_role", ["stream:owner", "stream:contributor", "stream:reviewer", None]
    )
    def test_project_update_role(
        self,
        client: SpeckleClient,
        second_client: SpeckleClient,
        project: Project,
        new_role: Optional[str],
        created_invite: PendingStreamCollaborator,
    ):
        assert created_invite.token

        input = ProjectInviteUseInput(
            accept=True, project_id=created_invite.projectId, token=created_invite.token
        )
        res = second_client.project_invite.use(input)

        invitee_id = second_client.account.userInfo.id
        assert invitee_id
        input = ProjectUpdateRoleInput(
            user_id=invitee_id,
            project_id=project.id,
            role=new_role,
        )
        res = client.project.update_role(input)

        assert isinstance(res, ProjectWithTeam)
        final_project = second_client.project.get(project.id)

        assert isinstance(res, Project)
        assert final_project.role == new_role



import os
import uuid
from typing import List, Optional, Tuple
from urllib.parse import urlparse

import pytest

from specklepy.core.api.credentials import Account, UserInfo, get_accounts_for_server
from specklepy.core.api.models import ServerInfo, ServerMigration
from specklepy.core.helpers import speckle_path_provider


def _create_account(
    id: str, url: str, movedFrom: Optional[str], movedTo: Optional[str]
) -> Account:
    return Account(
        id=uuid.uuid4().hex[:6].lower(),
        token="myToken",
        serverInfo=ServerInfo(
            url=url,
            name="myServer",
            migration=ServerMigration(moved_to=movedTo, moved_from=movedFrom),
        ),
        userInfo=UserInfo(id=id),
    )


def _test_cases() -> List[Tuple[List[Account], str, List[Account]]]:
    user_id_1 = uuid.uuid4().hex[:6].lower()
    user_id_2 = uuid.uuid4().hex[:6].lower()
    old = _create_account(
        user_id_1, "https://old.example.com", None, "https://new.example.com"
    )
    new = _create_account(
        user_id_1, "https://new.example.com", "https://old.example.com", None
    )
    other = _create_account(user_id_2, "https://other.example.com", None, None)

    given_accounts = [old, new, other]
    reversed = [other, new, old]

    return [
        (given_accounts, "https://old.example.com", [new]),
        (given_accounts, "https://new.example.com", [new]),
        (reversed, "https://old.example.com", [new]),
    ]


def _clean_accounts(accounts: List[Account]) -> None:
    json_accounts = speckle_path_provider.accounts_folder_path()

    for acc in accounts:
        # deleting acc json file in json_accounts path
        os.remove(os.path.join(json_accounts, f"{acc.id}.json"))
        pass


def _add_accounts(accounts: List[Account]) -> None:
    json_accounts = speckle_path_provider.accounts_folder_path()

    for acc in accounts:
        data = Account.model_dump_json(acc)
        with open(os.path.join(json_accounts, f"{acc.id}.json"), "w") as f:
            f.write(data)


@pytest.mark.parametrize("accounts, requested_url, expected", _test_cases())
def test_server_migration(
    accounts: List[Account], requested_url: str, expected: List[Account]
) -> None:
    _add_accounts(accounts)
    try:
        res = get_accounts_for_server(urlparse(requested_url).netloc)
        assert res == expected

    finally:
        _clean_accounts(accounts)



from typing import Optional

from specklepy.core.api.models import (
    LimitedUser,
    UserSearchResultCollection,
)
from specklepy.core.api.resources import OtherUserResource as CoreResource
from specklepy.logging import metrics


class OtherUserResource(CoreResource):
    """
    Provides API access to other users' profiles and activities on the platform.
    This class enables fetching limited information about users,
    searching for users by name or email,
    and accessing user activity logs with appropriate privacy
    and access control measures in place.
    """

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=(server_version,),
        )
        self.schema = LimitedUser

    def get(self, id: str) -> Optional[LimitedUser]:
        metrics.track(metrics.SDK, self.account, {"name": "Other User Get"})
        return super().get(id)

    def user_search(
        self,
        query: str,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        archived: bool = False,
        emailOnly: bool = False,
    ) -> UserSearchResultCollection:
        metrics.track(metrics.SDK, self.account, {"name": "Other User Search"})
        return super().user_search(
            query, limit=limit, cursor=cursor, archived=archived, emailOnly=emailOnly
        )



# following imports seem to be unnecessary, but they need to stay
# to not break the scripts using these functions as non-core
from specklepy.core.api.models import (
    LimitedUser,
    PendingStreamCollaborator,
    ServerInfo,
    User,
)

__all__ = [
    "LimitedUser",
    "PendingStreamCollaborator",
    "ServerInfo",
    "User",
]



from dataclasses import dataclass
from typing import List, Optional

from specklepy.objects.base import Base
from specklepy.objects.geometry.box import Box
from specklepy.objects.geometry.polyline import Polyline
from specklepy.objects.interfaces import ICurve, IHasArea, IHasUnits


@dataclass(kw_only=True)
class Curve(
    Base,
    ICurve,
    IHasArea,
    IHasUnits,
    speckle_type="Objects.Geometry.Curve",
    detachable={"points", "weights", "knots", "displayValue"},
    chunkable={"points": 31250, "weights": 31250, "knots": 31250},
):
    """
    a NURBS curve
    """

    degree: int
    periodic: bool
    rational: bool
    points: List[float]
    weights: List[float]
    knots: List[float]
    closed: bool
    displayValue: Polyline
    bbox: Optional[Box] = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"degree: {self.degree}, "
            f"periodic: {self.periodic}, "
            f"rational: {self.rational}, "
            f"closed: {self.closed}, "
            f"units: {self.units})"
        )

    @property
    def length(self) -> float:
        return self.__dict__.get("_length", 0.0)

    @length.setter
    def length(self, value: float) -> None:
        self.__dict__["_length"] = value

    @property
    def area(self) -> float:
        return self.__dict__.get("_area", 0.0)

    @area.setter
    def area(self, value: float) -> None:
        self.__dict__["_area"] = value



from typing import List, Tuple

import pytest

from specklepy.core.api.operations import deserialize, serialize
from specklepy.logging.exceptions import SpeckleException
from specklepy.objects.geometry import Line, Point, Polycurve, Polyline
from specklepy.objects.models.units import Units


@pytest.fixture
def sample_points() -> Tuple[Point, Point, Point]:
    return (
        Point(x=0.0, y=0.0, z=0.0, units=Units.m),
        Point(x=1.0, y=0.0, z=0.0, units=Units.m),
        Point(x=1.0, y=1.0, z=0.0, units=Units.m),
    )


@pytest.fixture
def sample_lines(sample_points: Tuple[Point, Point, Point]) -> List[Line]:
    p1, p2, p3 = sample_points
    return [
        Line(start=p1, end=p2, units=Units.m),
        Line(start=p2, end=p3, units=Units.m),
    ]


@pytest.fixture
def sample_polycurve(sample_lines: List[Line]) -> Polycurve:
    return Polycurve(segments=sample_lines, units=Units.m)


def test_polycurve_creation(sample_lines: List[Line]):
    polycurve = Polycurve(segments=sample_lines, units=Units.m)
    assert len(polycurve.segments) == 2
    assert polycurve.units == Units.m.value
    assert isinstance(polycurve.segments[0], Line)


def test_polycurve_is_closed(sample_points: Tuple[Point, Point, Point]):
    p1, p2, p3 = sample_points
    lines = [
        Line(start=p1, end=p2, units=Units.m),
        Line(start=p2, end=p3, units=Units.m),
        Line(start=p3, end=p1, units=Units.m),
    ]
    closed_polycurve = Polycurve(segments=lines, units=Units.m)
    assert closed_polycurve.is_closed()


def test_polycurve_not_closed(sample_polycurve: Polycurve):
    assert not sample_polycurve.is_closed()


@pytest.mark.parametrize("expected_length", [2.0])
def test_polycurve_length(sample_polycurve: Polycurve, expected_length: float):
    sample_polycurve.length = sample_polycurve.calculate_length()
    assert sample_polycurve.length == pytest.approx(expected_length)


@pytest.mark.parametrize(
    "points,expected_segments,expected_closed",
    [
        ([0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0], 2, False),
        (
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0],
            4,
            True,
        ),
    ],
)
def test_polycurve_from_polyline(
    points: List[float], expected_segments: int, expected_closed: bool
):
    polyline = Polyline(value=points, units=Units.m)
    polycurve = Polycurve.from_polyline(polyline)

    assert len(polycurve.segments) == expected_segments
    assert polycurve.units == Units.m.value
    assert isinstance(polycurve.segments[0], Line)
    assert polycurve.is_closed() == expected_closed


def test_polycurve_serialization(sample_polycurve: Polycurve):
    serialized = serialize(sample_polycurve)
    deserialized = deserialize(serialized)

    assert isinstance(deserialized, Polycurve)
    assert len(deserialized.segments) == len(sample_polycurve.segments)
    assert deserialized.units == sample_polycurve.units

    expectedSegment = sample_polycurve.segments[0]
    segment = deserialized.segments[0]
    assert isinstance(expectedSegment, Line)
    assert isinstance(segment, Line)

    assert segment.start.x == expectedSegment.start.x
    assert segment.start.y == expectedSegment.start.y
    assert segment.start.z == expectedSegment.start.z
    assert segment.end.x == expectedSegment.end.x
    assert segment.end.y == expectedSegment.end.y
    assert segment.end.z == expectedSegment.end.z


def test_polycurve_empty():
    polycurve = Polycurve(segments=[], units=Units.m)
    assert not polycurve.is_closed()
    assert polycurve.calculate_length() == 0.0


def test_polycurve_invalid_segment():
    with pytest.raises(SpeckleException):
        Polycurve(segments=["not a curve"], units=Units.m)



import pytest

from specklepy.api.client import SpeckleClient
from specklepy.core.api.inputs.project_inputs import ProjectCreateInput
from specklepy.core.api.inputs.user_inputs import UserProjectsFilter, UserUpdateInput
from specklepy.core.api.models import ResourceCollection, User
from specklepy.logging.exceptions import GraphQLException


@pytest.mark.run()
class TestActiveUserResource:
    def test_active_user_get(self, client: SpeckleClient):
        res = client.active_user.get()

        assert isinstance(res, User)

    def test_active_user_update(self, client: SpeckleClient):
        NEW_NAME = "Ron"
        NEW_BIO = "Now I have a bio, isn't that nice!"
        NEW_COMPANY = "Limited Cooperation Organization Inc"

        input = UserUpdateInput(name=NEW_NAME, bio=NEW_BIO, company=NEW_COMPANY)
        res = client.active_user.update(input=input)

        assert isinstance(res, User)
        assert res.name == NEW_NAME
        assert res.bio == NEW_BIO
        assert res.company == NEW_COMPANY

    def test_active_user_get_projects(self, client: SpeckleClient):
        existing = client.active_user.get_projects()

        p1 = client.project.create(
            ProjectCreateInput(name="Project 1", description=None, visibility=None)
        )
        p2 = client.project.create(
            ProjectCreateInput(name="Project 2", description=None, visibility=None)
        )

        res = client.active_user.get_projects()

        assert isinstance(res, ResourceCollection)
        assert len(res.items) == len(existing.items) + 2
        assert any(project.id == p1.id for project in res.items)
        assert any(project.id == p2.id for project in res.items)

    def test_active_user_get_projects_with_filter(self, client: SpeckleClient):
        # Since the client may be reused for other tests,
        # this test does rely on no other test creating a project
        # with "Search for me" in its name
        p1 = client.project.create(
            ProjectCreateInput(name="Search for me!", description=None, visibility=None)
        )
        _ = client.project.create(
            ProjectCreateInput(name="But not me!", description=None, visibility=None)
        )
        filter = UserProjectsFilter(search="Search for me")

        res = client.active_user.get_projects(filter=filter)

        assert isinstance(res, ResourceCollection)
        assert len(res.items) == 1
        assert res.total_count == 1
        assert res.items[0].id == p1.id

    def test_can_create_personal_projects(self, client: SpeckleClient):
        res = client.active_user.can_create_personal_projects()
        res.ensure_authorised()

        assert res.authorized is True

    def test_get_workspaces(self, client: SpeckleClient):
        """
        Test server is not workspace enabled, so we can't really test everything here
        We'll just test client's error handling
        """
        with pytest.raises(GraphQLException):
            _ = client.active_user.get_workspaces()

    def test_get_active_workspace(self, client: SpeckleClient):
        """
        Test server is not workspace enabled, so we can't really test everything here
        """
        res = client.active_user.get_active_workspace()
        assert res is None



import pytest

from specklepy.api.client import SpeckleClient
from specklepy.core.api.inputs.project_inputs import ProjectCreateInput
from specklepy.core.api.inputs.user_inputs import UserProjectsFilter
from specklepy.core.api.models.current import (
    Project,
    ProjectWithPermissions,
    ResourceCollection,
)


@pytest.mark.run()
class TestActiveUserResourcePermissions:
    @pytest.fixture()
    def test_project(self, client: SpeckleClient) -> Project:
        project = client.project.create(
            ProjectCreateInput(
                name="test project for active user permissions",
                description="test description",
                visibility=None,
            )
        )
        return project

    def test_active_user_get_projects_with_permissions(
        self, client: SpeckleClient, test_project: Project
    ):
        result = client.active_user.get_projects_with_permissions()

        assert isinstance(result, ResourceCollection)
        assert len(result.items) >= 1

        test_project_with_permissions = None
        for project in result.items:
            if project.id == test_project.id:
                test_project_with_permissions = project
                break

        assert test_project_with_permissions is not None
        assert isinstance(test_project_with_permissions, ProjectWithPermissions)

        assert hasattr(test_project_with_permissions, "permissions")
        assert test_project_with_permissions.permissions is not None

        assert test_project_with_permissions.id == test_project.id
        assert test_project_with_permissions.name == test_project.name

        permissions = test_project_with_permissions.permissions
        assert hasattr(permissions, "can_create_model")
        assert hasattr(permissions, "can_delete")
        assert hasattr(permissions, "can_load")
        assert hasattr(permissions, "can_publish")

        assert permissions.can_create_model.authorized is True
        assert permissions.can_delete.authorized is True
        assert permissions.can_load.authorized is True
        assert permissions.can_publish.authorized is True

    def test_active_user_get_projects_with_permissions_with_filter(
        self, client: SpeckleClient, test_project: Project
    ):
        """test getting active user's projects with permissions using a filter."""
        filter = UserProjectsFilter(search=test_project.name)

        result = client.active_user.get_projects_with_permissions(filter=filter)

        assert isinstance(result, ResourceCollection)
        assert len(result.items) >= 1
        assert result.total_count >= 1

        project_with_permissions = result.items[0]
        assert isinstance(project_with_permissions, ProjectWithPermissions)
        assert project_with_permissions.id == test_project.id

        assert hasattr(project_with_permissions, "permissions")
        assert project_with_permissions.permissions is not None

    def test_active_user_projects_with_permissions_method_exists(
        self, client: SpeckleClient
    ):
        """test that the method exists and is callable on active user resource."""
        assert hasattr(client.active_user, "get_projects_with_permissions")
        method = client.active_user.get_projects_with_permissions
        assert callable(method)



import pytest

from specklepy.core.api.operations import deserialize, serialize
from specklepy.objects.geometry import Point, PointCloud
from specklepy.objects.models.units import Units


@pytest.fixture
def sample_points():
    return [
        Point(x=0.0, y=0.0, z=0.0, units=Units.m),
        Point(x=1.0, y=0.0, z=0.0, units=Units.m),
        Point(x=0.0, y=1.0, z=0.0, units=Units.m),
        Point(x=1.0, y=1.0, z=0.0, units=Units.m),
    ]


@pytest.fixture
def sample_point_cloud(sample_points):
    return PointCloud(points=sample_points, units=Units.m)


def test_point_cloud_creation(sample_points):
    point_cloud = PointCloud(points=sample_points, units=Units.m)

    assert len(point_cloud.points) == 4
    assert isinstance(point_cloud.points, list)
    assert all(isinstance(p, Point) for p in point_cloud.points)
    assert point_cloud.units == Units.m.value


def test_point_cloud_units(sample_points):
    point_cloud = PointCloud(points=sample_points, units=Units.m)
    assert point_cloud.units == Units.m.value

    point_cloud.units = "mm"
    assert point_cloud.units == "mm"


def test_point_cloud_empty_points():
    point_cloud = PointCloud(points=[], units=Units.m)
    assert len(point_cloud.points) == 0
    assert isinstance(point_cloud.points, list)


def test_point_cloud_serialization(sample_point_cloud):
    serialized = serialize(sample_point_cloud)
    deserialized = deserialize(serialized)

    assert isinstance(deserialized, PointCloud)
    assert len(deserialized.points) == len(sample_point_cloud.points)

    for orig_point, deserial_point in zip(
        sample_point_cloud.points, deserialized.points, strict=True
    ):
        assert deserial_point.x == orig_point.x
        assert deserial_point.y == orig_point.y
        assert deserial_point.z == orig_point.z
        assert deserial_point.units == orig_point.units

    assert deserialized.units == sample_point_cloud.units



from typing import Optional

from gql import gql

from specklepy.core.api.inputs.model_inputs import (
    CreateModelInput,
    DeleteModelInput,
    ModelVersionsFilter,
    UpdateModelInput,
)
from specklepy.core.api.inputs.project_inputs import ProjectModelsFilter
from specklepy.core.api.models import Model, ModelWithVersions, ResourceCollection
from specklepy.core.api.resource import ResourceBase
from specklepy.core.api.responses import DataResponse

NAME = "model"


class ModelResource(ResourceBase):
    """API Access class for models"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )

    def get(self, model_id: str, project_id: str) -> Model:
        QUERY = gql(
            """
            query ModelGet($modelId: String!, $projectId: String!) {
              data:project(id: $projectId) {
                data:model(id: $modelId) {
                  id
                  name
                  previewUrl
                  updatedAt
                  description
                  displayName
                  createdAt
                  author {
                    avatar
                    bio
                    company
                    id
                    name
                    role
                    verified
                  }
                }
              }
            }
            """
        )

        variables = {
            "modelId": model_id,
            "projectId": project_id,
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[Model]], QUERY, variables
        ).data.data

    def get_with_versions(
        self,
        model_id: str,
        project_id: str,
        *,
        versions_limit: int = 25,
        versions_cursor: Optional[str] = None,
        versions_filter: Optional[ModelVersionsFilter] = None,
    ) -> ModelWithVersions:
        QUERY = gql(
            """
            query ModelGetWithVersions(
              $modelId: String!,
              $projectId: String!,
              $versionsLimit: Int!,
              $versionsCursor: String,
              $versionsFilter: ModelVersionsFilter
              ) {
              data:project(id: $projectId) {
                data:model(id: $modelId) {
                  id
                  name
                  previewUrl
                  updatedAt
                  versions(
                    limit: $versionsLimit,
                    cursor: $versionsCursor,
                    filter: $versionsFilter
                    ) {
                    items {
                      id
                      referencedObject
                      message
                      sourceApplication
                      createdAt
                      previewUrl
                      authorUser {
                        avatar
                        id
                        name
                        bio
                        company
                        verified
                        role
                      }
                    }
                    totalCount
                    cursor
                  }
                  description
                  displayName
                  createdAt
                  author {
                    avatar
                    bio
                    company
                    id
                    name
                    role
                    verified
                  }
                }
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
            "modelId": model_id,
            "versionsLimit": versions_limit,
            "versionsCursor": versions_cursor,
            "versionsFilter": (
                versions_filter.model_dump(warnings="error", by_alias=True)
                if versions_filter
                else None
            ),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[ModelWithVersions]], QUERY, variables
        ).data.data

    def get_models(
        self,
        project_id: str,
        *,
        models_limit: int = 25,
        models_cursor: Optional[str] = None,
        models_filter: Optional[ProjectModelsFilter] = None,
    ) -> ResourceCollection[Model]:
        QUERY = gql(
            """
            query ProjectGetWithModels(
              $projectId: String!,
              $modelsLimit: Int!,
              $modelsCursor: String,
              $modelsFilter: ProjectModelsFilter
              ) {
              data:project(id: $projectId) {
                data:models(
                  limit: $modelsLimit,
                  cursor: $modelsCursor,
                  filter: $modelsFilter
                  ) {
                  items {
                    id
                    name
                    previewUrl
                    updatedAt
                    displayName
                    description
                    createdAt
                    author {
                      avatar
                      bio
                      company
                      id
                      name
                      role
                      verified
                    }
                  }
                  totalCount
                  cursor
                }
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
            "modelsLimit": models_limit,
            "modelsCursor": models_cursor,
            "modelsFilter": (
                models_filter.model_dump(warnings="error", by_alias=True)
                if models_filter
                else None
            ),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[ResourceCollection[Model]]], QUERY, variables
        ).data.data

    def create(self, input: CreateModelInput) -> Model:
        QUERY = gql(
            """
            mutation ModelCreate($input: CreateModelInput!) {
              data:modelMutations {
                data:create(input: $input) {
                  id
                  displayName
                  name
                  description
                  createdAt
                  updatedAt
                  previewUrl
                  author {
                    avatar
                    bio
                    company
                    id
                    name
                    role
                    verified
                  }
                }
              }
            }
            """
        )

        variables = {
            "input": input.model_dump(warnings="error", by_alias=True),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[Model]], QUERY, variables
        ).data.data

    def delete(self, input: DeleteModelInput) -> bool:
        QUERY = gql(
            """
            mutation ModelDelete($input: DeleteModelInput!) {
              data:modelMutations {
                data:delete(input: $input)
              }
            }
            """
        )

        variables = {"input": input.model_dump(warnings="error", by_alias=True)}

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[bool]], QUERY, variables
        ).data.data

    def update(self, input: UpdateModelInput) -> Model:
        QUERY = gql(
            """
            mutation ModelUpdate($input: UpdateModelInput!) {
              data:modelMutations {
                data:update(input: $input) {
                  id
                  name
                  displayName
                  description
                  createdAt
                  updatedAt
                  previewUrl
                  author {
                    avatar
                    bio
                    company
                    id
                    name
                    role
                    verified
                  }
                }
              }
            }
            """
        )

        variables = {
            "input": input.model_dump(warnings="error", by_alias=True),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[Model]], QUERY, variables
        ).data.data



from dataclasses import dataclass

from specklepy.objects.base import Base
from specklepy.objects.geometry.plane import Plane
from specklepy.objects.interfaces import ICurve, IHasArea, IHasUnits


@dataclass(kw_only=True)
class Ellipse(
    Base, IHasUnits, ICurve, IHasArea, speckle_type="Objects.Geometry.Ellipse"
):
    """
    an ellipse
    """

    plane: Plane
    first_radius: float
    second_radius: float

    @property
    def length(self) -> float:
        return self.__dict__.get("_length", 0.0)

    @length.setter
    def length(self, value: float) -> None:
        self.__dict__["_length"] = value

    @property
    def area(self) -> float:
        return self.__dict__.get("_area", 0.0)

    @area.setter
    def area(self, value: float) -> None:
        self.__dict__["_area"] = value



from typing import List, Optional

from specklepy.core.api.inputs.user_inputs import (
    UserProjectsFilter,
    UserUpdateInput,
    UserWorkspacesFilter,
)
from specklepy.core.api.models import (
    PendingStreamCollaborator,
    Project,
    ResourceCollection,
    User,
)
from specklepy.core.api.models.current import (
    LimitedWorkspace,
    PermissionCheckResult,
    ProjectWithPermissions,
    Workspace,
)
from specklepy.core.api.resources import ActiveUserResource as CoreResource
from specklepy.logging import metrics


class ActiveUserResource(CoreResource):
    """API Access class for users. This class provides methods to get and update
    the user profile, fetch user activity, and manage pending stream invitations."""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )
        self.schema = User

    def get(self) -> Optional[User]:
        metrics.track(metrics.SDK, self.account, {"name": "Active User Get"})
        return super().get()

    def update(
        self,
        input: UserUpdateInput,
    ) -> User:
        metrics.track(metrics.SDK, self.account, {"name": "Active User Update"})

        return super().update(input=input)

    def get_projects(
        self,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        filter: Optional[UserProjectsFilter] = None,
    ) -> ResourceCollection[Project]:
        metrics.track(metrics.SDK, self.account, {"name": "Active User Get Projects"})
        return super().get_projects(limit=limit, cursor=cursor, filter=filter)

    def get_projects_with_permissions(
        self,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        filter: Optional[UserProjectsFilter] = None,
    ) -> ResourceCollection[ProjectWithPermissions]:
        metrics.track(
            metrics.SDK,
            self.account,
            {"name": "Active User Get Projects With Permissions"},
        )
        return super().get_projects_with_permissions(
            limit=limit, cursor=cursor, filter=filter
        )

    def get_project_invites(self) -> List[PendingStreamCollaborator]:
        metrics.track(
            metrics.SDK, self.account, {"name": "Active User Get Project Invites"}
        )
        return super().get_project_invites()

    def can_create_personal_projects(self) -> PermissionCheckResult:
        metrics.track(
            metrics.SDK,
            self.account,
            {"name": "Active User Can Create Personal Projects Check"},
        )
        return super().can_create_personal_projects()

    def get_workspaces(
        self,
        limit: int = 25,
        cursor: Optional[str] = None,
        filter: Optional[UserWorkspacesFilter] = None,
    ) -> ResourceCollection[Workspace]:
        metrics.track(metrics.SDK, self.account, {"name": "Active User Get Workspaces"})
        return super().get_workspaces(limit, cursor, filter)

    def get_active_workspace(self) -> Optional[LimitedWorkspace]:
        metrics.track(
            metrics.SDK, self.account, {"name": "Active User Get Active Workspace"}
        )
        return super().get_active_workspace()



import pytest

from specklepy.api.client import SpeckleClient
from specklepy.core.api.inputs.model_inputs import (
    CreateModelInput,
    DeleteModelInput,
    UpdateModelInput,
)
from specklepy.core.api.inputs.project_inputs import (
    ProjectCreateInput,
    ProjectModelsFilter,
)
from specklepy.core.api.models.current import (
    Model,
    Project,
    ProjectWithModels,
    ResourceCollection,
)
from specklepy.logging.exceptions import GraphQLException


@pytest.mark.run()
class TestModelResource:
    @pytest.fixture()
    def test_project(self, client: SpeckleClient) -> Project:
        project = client.project.create(
            ProjectCreateInput(name="Test project", description="", visibility=None)
        )
        return project

    @pytest.fixture()
    def test_model(self, client: SpeckleClient, test_project: Project) -> Model:
        model = client.model.create(
            CreateModelInput(
                name="Test Model", description="", project_id=test_project.id
            )
        )
        return model

    @pytest.mark.parametrize(
        "name, description",
        [
            ("My Model", "My model description"),
            ("my/nested/model", None),
        ],
    )
    def test_model_create(
        self, client: SpeckleClient, test_project: Project, name: str, description: str
    ):
        input = CreateModelInput(
            name=name, description=description, project_id=test_project.id
        )
        result = client.model.create(input)

        assert isinstance(result, Model)
        assert result.name.lower() == name.lower()
        assert result.description == description

    def test_model_get(
        self, client: SpeckleClient, test_model: Model, test_project: Project
    ):
        result = client.model.get(test_model.id, test_project.id)

        assert isinstance(result, Model)
        assert result.id == test_model.id
        assert result.name == test_model.name
        assert result.description == test_model.description
        assert result.created_at == test_model.created_at
        assert result.updated_at == test_model.updated_at

    def test_models_get_with_filter(
        self, client: SpeckleClient, test_model: Model, test_project: Project
    ):
        filter = ProjectModelsFilter(search=test_model.name)

        result = client.model.get_models(test_project.id, models_filter=filter)

        assert isinstance(result, ResourceCollection)
        assert len(result.items) == 1
        assert result.total_count == 1
        assert result.items[0].id == test_model.id

    def test_get_models(
        self, client: SpeckleClient, test_project: Project, test_model: Model
    ):
        result = client.model.get_models(test_project.id)

        assert isinstance(result, ResourceCollection)
        assert len(result.items) == 1
        assert result.total_count == 1
        assert result.items[0].id == test_model.id

    def test_project_get_models(
        self, client: SpeckleClient, test_project: Project, test_model: Model
    ):
        result = client.project.get_with_models(test_project.id)

        assert isinstance(result, ProjectWithModels)
        assert result.id == test_project.id
        assert isinstance(result.models, ResourceCollection)
        assert len(result.models.items) == 1
        assert result.models.total_count == 1
        assert result.models.items[0].id == test_model.id

    def test_project_get_models_with_filter(
        self, client: SpeckleClient, test_project: Project, test_model: Model
    ):
        filter = ProjectModelsFilter(search=test_model.name)
        result = client.project.get_with_models(test_project.id, models_filter=filter)

        assert isinstance(result, ProjectWithModels)
        assert result.id == test_project.id
        assert isinstance(result.models, ResourceCollection)
        assert len(result.models.items) == 1
        assert result.models.total_count == 1
        assert result.models.items[0].id == test_model.id

    def test_model_update(
        self, client: SpeckleClient, test_model: Model, test_project: Project
    ):
        new_name = "MY new name"
        new_description = "MY new desc"

        update_data = UpdateModelInput(
            id=test_model.id,
            name=new_name,
            description=new_description,
            project_id=test_project.id,
        )

        updated_model = client.model.update(update_data)

        assert isinstance(updated_model, Model)
        assert updated_model.id == test_model.id
        assert updated_model.name.lower() == new_name.lower()
        assert updated_model.description == new_description
        assert updated_model.updated_at >= test_model.updated_at

    def test_model_delete(
        self, client: SpeckleClient, test_model: Model, test_project: Project
    ):
        delete_data = DeleteModelInput(id=test_model.id, project_id=test_project.id)

        response = client.model.delete(delete_data)
        assert response is True

        with pytest.raises(GraphQLException):
            client.model.get(test_model.id, test_project.id)

        with pytest.raises(GraphQLException):
            client.model.delete(delete_data)



import pytest

from specklepy.api.client import SpeckleClient
from specklepy.core.api.inputs.model_inputs import CreateModelInput, ModelVersionsFilter
from specklepy.core.api.inputs.project_inputs import ProjectCreateInput
from specklepy.core.api.inputs.version_inputs import (
    DeleteVersionsInput,
    MarkReceivedVersionInput,
    MoveVersionsInput,
    UpdateVersionInput,
)
from specklepy.core.api.models import (
    Model,
    ModelWithVersions,
    Project,
    ResourceCollection,
    Version,
)
from specklepy.logging.exceptions import GraphQLException
from tests.integration.conftest import create_version


@pytest.mark.run()
class TestVersionResource:
    @pytest.fixture
    def test_project(self, client: SpeckleClient) -> Project:
        project = client.project.create(
            ProjectCreateInput(name="Test project", description="", visibility=None)
        )
        return project

    @pytest.fixture
    def test_model_1(self, client: SpeckleClient, test_project: Project) -> Model:
        model1 = client.model.create(
            CreateModelInput(
                name="Test Model 1", description="", project_id=test_project.id
            )
        )
        return model1

    @pytest.fixture
    def test_model_2(self, client: SpeckleClient, test_project: Project) -> Model:
        model2 = client.model.create(
            CreateModelInput(
                name="Test Model 2", description="", project_id=test_project.id
            )
        )
        return model2

    @pytest.fixture
    def test_version(
        self, client: SpeckleClient, test_project: Project, test_model_1: Model
    ) -> Version:
        return create_version(client, test_project.id, test_model_1.id)

    def test_version_get(
        self, client: SpeckleClient, test_version: Version, test_project: Project
    ):
        result = client.version.get(test_version.id, test_project.id)

        assert isinstance(result, Version)
        assert result.id == test_version.id
        assert result.message == test_version.message

    def test_versions_get(
        self,
        client: SpeckleClient,
        test_model_1: Model,
        test_project: Project,
        test_version: Version,
    ):
        result = client.version.get_versions(test_model_1.id, test_project.id)

        assert isinstance(result, ResourceCollection)
        assert len(result.items) == 1
        assert result.total_count == 1
        assert result.items[0].id == test_version.id

    def test_versions_get_with_filter(
        self,
        client: SpeckleClient,
        test_model_1: Model,
        test_project: Project,
        test_version: Version,
    ):
        filter = ModelVersionsFilter(
            priority_ids=[test_version.id], priority_ids_only=True
        )

        result = client.version.get_versions(
            test_model_1.id, test_project.id, filter=filter
        )

        assert isinstance(result, ResourceCollection)
        assert len(result.items) == 1
        assert result.total_count == 1
        assert result.items[0].id == test_version.id

    def test_version_received(
        self, client: SpeckleClient, test_version: Version, test_project: Project
    ):
        input = MarkReceivedVersionInput(
            version_id=test_version.id,
            project_id=test_project.id,
            source_application="Integration test",
        )
        result = client.version.received(input)

        assert result is True

    def test_model_get_with_versions(
        self,
        client: SpeckleClient,
        test_model_1: Model,
        test_project: Project,
        test_version: Version,
    ):
        result = client.model.get_with_versions(test_model_1.id, test_project.id)

        assert isinstance(result, ModelWithVersions)
        assert result.id == test_model_1.id
        assert len(result.versions.items) == 1
        assert result.versions.total_count == 1
        assert result.versions.items[0].id == test_version.id

    def test_model_get_with_versions_with_filter(
        self,
        client: SpeckleClient,
        test_model_1: Model,
        test_project: Project,
        test_version: Version,
    ):
        filter = ModelVersionsFilter(
            priority_ids=[test_version.id], priority_ids_only=True
        )

        result = client.model.get_with_versions(
            test_model_1.id, test_project.id, versions_filter=filter
        )

        assert isinstance(result, ModelWithVersions)
        assert len(result.versions.items) == 1
        assert result.versions.total_count == 1
        assert isinstance(result.versions, ResourceCollection)
        assert result.versions.items[0].id == test_version.id

    def test_version_update(
        self, client: SpeckleClient, test_version: Version, test_project: Project
    ):
        new_message = "MY new version message"
        input = UpdateVersionInput(
            version_id=test_version.id, project_id=test_project.id, message=new_message
        )
        updated_version = client.version.update(input)

        assert isinstance(updated_version, Version)
        assert updated_version.id == test_version.id
        assert updated_version.message == new_message
        assert updated_version.preview_url == test_version.preview_url

    def test_version_move_to_model(
        self,
        client: SpeckleClient,
        test_project: Project,
        test_version: Version,
        test_model_2: Model,
    ):
        input = MoveVersionsInput(
            target_model_name=test_model_2.name,
            version_ids=[test_version.id],
            project_id=test_project.id,
        )
        moved_model_id = client.version.move_to_model(input)

        assert isinstance(moved_model_id, str)
        assert moved_model_id == test_model_2.id
        moved_version = client.version.get(test_version.id, test_project.id)

        assert isinstance(moved_version, Version)
        assert moved_version.id == test_version.id
        assert moved_version.message == test_version.message
        assert moved_version.preview_url == test_version.preview_url

    def test_version_delete(
        self, client: SpeckleClient, test_version: Version, test_project: Project
    ):
        input = DeleteVersionsInput(
            version_ids=[test_version.id], project_id=test_project.id
        )

        response = client.version.delete(input)
        assert response is True

        with pytest.raises(GraphQLException):
            client.version.get(test_version.id, test_project.id)

        with pytest.raises(GraphQLException):
            client.version.delete(input)



import os
from pathlib import Path
from typing import List, Optional
from urllib.parse import urlparse

from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module

from specklepy.core.api.models import ServerInfo
from specklepy.core.helpers import speckle_path_provider
from specklepy.logging.exceptions import SpeckleException
from specklepy.transports.sqlite import SQLiteTransport


class UserInfo(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    email: Optional[str] = None
    company: Optional[str] = None
    avatar: Optional[str] = None


class Account(BaseModel):
    isDefault: bool = False
    token: Optional[str] = None
    refreshToken: Optional[str] = None
    serverInfo: ServerInfo = Field(default_factory=ServerInfo)
    userInfo: UserInfo = Field(default_factory=UserInfo)
    id: Optional[str] = None

    def __repr__(self) -> str:
        return (
            f"Account(email: {self.userInfo.email}, server: {self.serverInfo.url},"
            f" isDefault: {self.isDefault})"
        )

    def __str__(self) -> str:
        return self.__repr__()

    @classmethod
    def from_token(cls, token: str, server_url: str = None):
        acct = cls(token=token)
        acct.serverInfo.url = server_url
        return acct


def get_local_accounts(base_path: Optional[str] = None) -> List[Account]:
    """Gets all the accounts present in this environment

    Arguments:
        base_path {str} -- custom base path if you are not using the system default

    Returns:
        List[Account] -- list of all local accounts or an empty list if
        no accounts were found
    """
    accounts: List[Account] = []
    try:
        account_storage = SQLiteTransport(scope="Accounts", base_path=base_path)
        res = account_storage.get_all_objects()
        account_storage.close()
        if res:
            accounts.extend(Account.model_validate_json(r[1]) for r in res)
    except SpeckleException:
        # cannot open SQLiteTransport, probably because of the lack
        # of disk write permissions
        pass

    json_acct_files = []
    json_path = str(speckle_path_provider.accounts_folder_path())
    try:
        os.makedirs(json_path, exist_ok=True)
        json_acct_files.extend(
            file for file in os.listdir(json_path) if file.endswith(".json")
        )

    except Exception:
        # cannot find or get the json account paths
        pass

    if json_acct_files:
        try:
            accounts.extend(
                Account.model_validate_json(Path(json_path, json_file).read_text())
                # Account.parse_file(os.path.join(json_path, json_file))
                for json_file in json_acct_files
            )
        except Exception as ex:
            raise SpeckleException(
                "Invalid json accounts could not be read. Please fix or remove them.",
                ex,
            ) from ex

    return accounts


def get_default_account(base_path: Optional[str] = None) -> Optional[Account]:
    """
    Gets this environment's default account if any. If there is no default,
    the first found will be returned and set as default.
    Arguments:
        base_path {str} -- custom base path if you are not using the system default

    Returns:
        Account -- the default account or None if no local accounts were found
    """
    accounts = get_local_accounts(base_path=base_path)
    if not accounts:
        return None

    default = next((acc for acc in accounts if acc.isDefault), None)
    if not default:
        default = accounts[0]
        default.isDefault = True
    # metrics.initialise_tracker(default)

    return default


def get_account_from_token(token: str, server_url: str = None) -> Account:
    """Gets the local account for the token if it exists
    Arguments:
        token {str} -- the api token

    Returns:
        Account -- the local account with this token or a shell account containing
        just the token and url if no local account is found
    """
    accounts = get_local_accounts()
    if not accounts:
        return Account.from_token(token, server_url)

    acct = next((acc for acc in accounts if acc.token == token), None)
    if acct:
        return acct

    if server_url:
        url = server_url.lower()
        acct = next(
            (acc for acc in accounts if url in acc.serverInfo.url.lower()), None
        )
        if acct:
            return acct

    return Account.from_token(token, server_url)


def get_accounts_for_server(host: str) -> List[Account]:
    all_accounts = get_local_accounts()
    filtered: List[Account] = []

    for acc in all_accounts:
        moved_from = (
            acc.serverInfo.migration.moved_from if acc.serverInfo.migration else None
        )

        if moved_from and host == urlparse(moved_from).netloc:
            filtered.append(acc)

    for acc in all_accounts:
        if any([x for x in filtered if x.userInfo.id == acc.userInfo.id]):
            continue

        if host == urlparse(acc.serverInfo.url).netloc:
            filtered.append(acc)

    return filtered


class StreamWrapper:
    def __init__(self, url: str = None) -> None:
        raise SpeckleException(
            message=(
                "The StreamWrapper has moved as of v2.6.0! Please import from"
                " specklepy.api.wrapper"
            ),
            exception=DeprecationWarning(),
        )



import pytest
from httpx import HTTPStatusError

from specklepy.core.api.connector_versions import (
    ConnectorVersion,
    ConnectorVersions,
    get_connector_versions,
    get_latest_version,
)

# NOTE: the tests in this file are testing against the live releases.speckle.dev server
# url defined in get_connector_versions.


def test_connector_versions():
    res = get_connector_versions("blender")

    assert isinstance(res, ConnectorVersions)
    assert res.versions  # Assuming the feed is not empty


def test_get_latest_version_throws_no_slug():
    with pytest.raises(HTTPStatusError) as ex:
        get_latest_version("non-existent-connector!", True)

    assert "404" in str(ex.value)


def test_get_latest_version():
    res = get_latest_version("blender", False)

    assert isinstance(res, ConnectorVersion)



from dataclasses import dataclass

from specklepy.objects.base import Base
from specklepy.objects.geometry.point import Point
from specklepy.objects.geometry.vector import Vector
from specklepy.objects.interfaces import IHasUnits


@dataclass(kw_only=True)
class Plane(Base, IHasUnits, speckle_type="Objects.Geometry.Plane"):
    """
    a plane consisting of an origin Point, and 3 Vectors as its X, Y and Z axis.
    """

    origin: Point
    normal: Vector
    xdir: Vector
    ydir: Vector

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"origin: {self.origin}, "
            f"normal: {self.normal}, "
            f"xdir: {self.xdir}, "
            f"ydir: {self.ydir}, "
            f"units: {self.units})"
        )



from typing import Optional, Sequence

from specklepy.core.api.enums import ProjectVisibility
from specklepy.core.api.models.graphql_base_model import GraphQLBaseModel


class ProjectCreateInput(GraphQLBaseModel):
    name: Optional[str]
    description: Optional[str]
    visibility: Optional[ProjectVisibility]


class WorkspaceProjectCreateInput(GraphQLBaseModel):
    name: Optional[str]
    description: Optional[str]
    visibility: Optional[ProjectVisibility]
    workspaceId: str


class ProjectInviteCreateInput(GraphQLBaseModel):
    email: Optional[str]
    role: Optional[str]
    server_role: Optional[str]
    userId: Optional[str]


class ProjectInviteUseInput(GraphQLBaseModel):
    accept: bool
    project_id: str
    token: str


class ProjectModelsFilter(GraphQLBaseModel):
    contributors: Optional[Sequence[str]] = None
    exclude_ids: Optional[Sequence[str]] = None
    ids: Optional[Sequence[str]] = None
    only_with_versions: Optional[bool] = None
    search: Optional[str] = None
    source_apps: Optional[Sequence[str]] = None


class ProjectUpdateInput(GraphQLBaseModel):
    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    allow_public_comments: Optional[bool] = None
    visibility: Optional[ProjectVisibility] = None


class ProjectUpdateRoleInput(GraphQLBaseModel):
    user_id: str
    project_id: str
    role: Optional[str]


class WorksaceProjectsFilter(GraphQLBaseModel):
    search: Optional[str]
    """Filter out projects by name"""
    with_project_role_only: Optional[bool]
    """
    Only return workspace projects that the active user has an explicit project role in
    """



import json
from typing import Dict, List, Optional
from warnings import warn

import requests

from specklepy.core.api.client import SpeckleClient
from specklepy.core.api.credentials import Account, get_account_from_token
from specklepy.logging.exceptions import SpeckleException, SpeckleWarning
from specklepy.transports.abstract_transport import AbstractTransport

from .batch_sender import BatchSender


class ServerTransport(AbstractTransport):
    """
    The `ServerTransport` is the vehicle through which you transport objects to and
    from a Speckle Server. Provide it to `operations.send()` or `operations.receive()`.

    The `ServerTransport` can be authenticated two different ways:
        1. by providing a `SpeckleClient`
        2. by providing an `Account`
        3. by providing a `token` and `url`

    ```py
    from specklepy.api import operations
    from specklepy.api.client import SpeckleClient
    from specklepy.core.api.inputs.version_inputs import CreateVersionInput
    from specklepy.transports.server import ServerTransport

    # here's the data you want to send
    block = Block(length=2, height=4)

    # here's the project and model you want to send to
    project_id = "abcdefghi"
    model_id = "ihgfedcba"

    # next, create the server transport - this is the vehicle through which
    # you will send and receive
    transport = ServerTransport(stream_id=project_id, client=client)

    # this serialises the block and sends it to the transport
    hash = operations.send(base=block, transports=[transport])

    # you can now create tag this version of the model with this object
    input = CreateVersionInput(
        objectId = hash,
        modelId = model_id,
        projectId = project_id,
    )
    version = client.version.create(input)
    ```
    """

    def __init__(
        self,
        stream_id: str,
        client: Optional[SpeckleClient] = None,
        account: Optional[Account] = None,
        token: Optional[str] = None,
        url: Optional[str] = None,
        name: str = "RemoteTransport",
    ) -> None:
        super().__init__()
        if client is None and account is None and token is None and url is None:
            raise SpeckleException(
                "You must provide either a client or a token and url to construct a"
                " ServerTransport."
            )

        self._name = name
        self.account = None
        self.saved_obj_count = 0
        if account:
            self.account = account
            url = account.serverInfo.url
        elif client:
            url = client.url
            if not client.account.token:
                warn(
                    SpeckleWarning(
                        "Unauthenticated Speckle Client provided to Server Transport"
                        f" for {url}. Receiving from private streams will fail."
                    ),
                    stacklevel=2,
                )
            else:
                self.account = client.account
        else:
            self.account = get_account_from_token(token, url)

        self.stream_id = stream_id
        self.url = url

        self.session = requests.Session()

        self.session.headers.update(
            {
                "Accept": "text/plain",
            }
        )

        if self.account is not None:
            self._batch_sender = BatchSender(
                self.url, self.stream_id, self.account.token, max_batch_size_mb=1
            )
            self.session.headers.update(
                {
                    "Authorization": f"Bearer {self.account.token}",
                }
            )

    @property
    def name(self) -> str:
        return self._name

    def begin_write(self) -> None:
        self.saved_obj_count = 0

    def end_write(self) -> None:
        self._batch_sender.flush()

    def save_object(self, id: str, serialized_object: str) -> None:
        self._batch_sender.send_object(id, serialized_object)

    def save_object_from_transport(
        self, id: str, source_transport: AbstractTransport
    ) -> None:
        obj_string = source_transport.get_object(id=id)
        self.save_object(id=id, serialized_object=obj_string)

    def get_object(self, id: str) -> str:
        # endpoint = f"{self.url}/objects/{self.stream_id}/{id}/single"
        # r = self.session.get(endpoint, stream=True)

        # _, obj = next(r.iter_lines().decode("utf-8")).split("\t")

        # return obj

        raise SpeckleException(
            "Getting a single object using `ServerTransport.get_object()` is not"
            " implemented.",
            NotImplementedError(),
        )

    def has_objects(self, id_list: List[str]) -> Dict[str, bool]:
        return {id: False for id in id_list}

    def copy_object_and_children(
        self, id: str, target_transport: AbstractTransport
    ) -> str:
        endpoint = f"{self.url}/objects/{self.stream_id}/{id}/single"
        r = self.session.get(endpoint)
        r.encoding = "utf-8"

        if r.status_code != 200:
            raise SpeckleException(
                f"Can't get object {self.stream_id}/{id}: HTTP error"
                f" {r.status_code} ({r.text[:1000]})"
            )
        root_obj_serialized = r.text
        root_obj = json.loads(root_obj_serialized)
        closures = root_obj.get("__closure", {})

        # Check which children are not already in the target transport
        children_ids = list(closures.keys())
        children_found_map = target_transport.has_objects(children_ids)
        new_children_ids = [
            id for id in children_found_map if not children_found_map[id]
        ]

        # Get the new children
        endpoint = f"{self.url}/api/getobjects/{self.stream_id}"
        r = self.session.post(
            endpoint, data={"objects": json.dumps(new_children_ids)}, stream=True
        )
        r.encoding = "utf-8"
        lines = r.iter_lines(decode_unicode=True)

        # iter through returned objects saving them as we go
        target_transport.begin_write()
        for line in lines:
            if line:
                hash, obj = line.split("\t")
                target_transport.save_object(hash, obj)

        target_transport.save_object(id, root_obj_serialized)
        target_transport.end_write()

        return root_obj_serialized



from typing import Optional

from specklepy.core.api.inputs.model_inputs import (
    CreateModelInput,
    DeleteModelInput,
    ModelVersionsFilter,
    UpdateModelInput,
)
from specklepy.core.api.inputs.project_inputs import ProjectModelsFilter
from specklepy.core.api.models import Model, ModelWithVersions, ResourceCollection
from specklepy.core.api.resources import ModelResource as CoreResource
from specklepy.logging import metrics


class ModelResource(CoreResource):
    """API Access class for models"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )

    def get(self, model_id: str, project_id: str) -> Model:
        metrics.track(metrics.SDK, self.account, {"name": "Model Get"})
        return super().get(model_id, project_id)

    def get_with_versions(
        self,
        model_id: str,
        project_id: str,
        *,
        versions_limit: int = 25,
        versions_cursor: Optional[str] = None,
        versions_filter: Optional[ModelVersionsFilter] = None,
    ) -> ModelWithVersions:
        metrics.track(metrics.SDK, self.account, {"name": "Model Get With Versions"})
        return super().get_with_versions(
            model_id,
            project_id,
            versions_limit=versions_limit,
            versions_cursor=versions_cursor,
            versions_filter=versions_filter,
        )

    def get_models(
        self,
        project_id: str,
        *,
        models_limit: int = 25,
        models_cursor: Optional[str] = None,
        models_filter: Optional[ProjectModelsFilter] = None,
    ) -> ResourceCollection[Model]:
        metrics.track(metrics.SDK, self.account, {"name": "Model Get Models"})
        return super().get_models(
            project_id,
            models_limit=models_limit,
            models_cursor=models_cursor,
            models_filter=models_filter,
        )

    def create(self, input: CreateModelInput) -> Model:
        metrics.track(metrics.SDK, self.account, {"name": "Model Create"})
        return super().create(input)

    def delete(self, input: DeleteModelInput) -> bool:
        metrics.track(metrics.SDK, self.account, {"name": "Model Delete"})
        return super().delete(input)

    def update(self, input: UpdateModelInput) -> Model:
        metrics.track(metrics.SDK, self.account, {"name": "Model Update"})
        return super().update(input)



# ignoring "line too long" check from linter
# to match with SpeckleExceptions
# ruff: noqa: E501

import pytest

from specklepy.core.api.operations import deserialize, serialize
from specklepy.logging.exceptions import SpeckleException
from specklepy.objects.geometry import ControlPoint
from specklepy.objects.models.units import Units


@pytest.fixture
def sample_control_point():
    return ControlPoint(x=1.0, y=2.0, z=3.0, weight=1.0, units=Units.m)


@pytest.mark.parametrize(
    "x,y,z,weight,expected_units", [(1.0, 2.0, 3.0, 1.0, Units.m.value)]
)
def test_control_point_creation(x, y, z, weight, expected_units):
    control_point = ControlPoint(x=x, y=y, z=z, weight=weight, units=Units.m)
    assert control_point.x == x
    assert control_point.y == y
    assert control_point.z == z
    assert control_point.weight == weight
    assert control_point.units == expected_units


@pytest.mark.parametrize("new_units", ["mm", "cm", "in"])
def test_control_point_units(new_units):
    control_point = ControlPoint(x=1.0, y=2.0, z=3.0, weight=1.0, units=Units.m)
    assert control_point.units == Units.m.value
    control_point.units = new_units
    assert control_point.units == new_units


@pytest.mark.parametrize(
    "invalid_param, test_params, error_msg",
    [
        (
            "x",
            {"x": "not a number", "y": 2.0, "z": 3.0, "weight": 1.0},
            "Cannot set 'ControlPoint.x':it expects type '<class 'float'>',but received type 'str'",
        ),
        (
            "y",
            {"x": 1.0, "y": "not a number", "z": 3.0, "weight": 1.0},
            "Cannot set 'ControlPoint.y':it expects type '<class 'float'>',but received type 'str'",
        ),
        (
            "z",
            {"x": 1.0, "y": 2.0, "z": "not a number", "weight": 1.0},
            "Cannot set 'ControlPoint.z':it expects type '<class 'float'>',but received type 'str'",
        ),
        (
            "weight",
            {"x": 1.0, "y": 2.0, "z": 3.0, "weight": "not a number"},
            "Cannot set 'ControlPoint.weight':it expects type '<class 'float'>',but received type 'str'",
        ),
    ],
)
def test_control_point_invalid_construction(invalid_param, test_params, error_msg):
    with pytest.raises(SpeckleException) as exc_info:
        ControlPoint(**test_params, units=Units.m)
    assert str(exc_info.value) == f"SpeckleException: {error_msg}"


def test_control_point_serialization(sample_control_point):
    serialized = serialize(sample_control_point)
    deserialized = deserialize(serialized)

    assert isinstance(deserialized, ControlPoint)
    assert deserialized.x == sample_control_point.x
    assert deserialized.y == sample_control_point.y
    assert deserialized.z == sample_control_point.z
    assert deserialized.weight == sample_control_point.weight
    assert deserialized.units == sample_control_point.units



import pytest

from specklepy.core.api.operations import deserialize, serialize
from specklepy.objects.geometry import Point
from specklepy.objects.models.units import Units


def test_point_creation():
    p1 = Point(x=1.0, y=2.0, z=3.0, units=Units.m)
    assert p1.x == 1.0
    assert p1.y == 2.0
    assert p1.z == 3.0
    assert p1.units == Units.m.value


def test_point_distance_calculation():
    p1 = Point(x=1.0, y=2.0, z=3.0, units=Units.m)
    p2 = Point(x=4.0, y=6.0, z=8.0, units=Units.m)

    distance = p1.distance_to(p2)
    expected = (3.0**2 + 4.0**2 + 5.0**2) ** 0.5
    assert distance == pytest.approx(expected)

    with pytest.raises(TypeError):
        p1.distance_to("not a point")


def test_point_serialization():
    p1 = Point(x=1.0, y=2.0, z=3.0, units=Units.mm)
    serialized = serialize(p1)
    deserialized = deserialize(serialized)

    assert isinstance(deserialized, Point)
    assert deserialized.x == p1.x
    assert deserialized.y == p1.y
    assert deserialized.z == p1.z
    assert deserialized.units == p1.units



from datetime import datetime
from typing import Generic, List, Optional, TypeVar

from specklepy.core.api.enums import ProjectVisibility
from specklepy.core.api.models.graphql_base_model import GraphQLBaseModel
from specklepy.logging.exceptions import WorkspacePermissionException

T = TypeVar("T")


class User(GraphQLBaseModel):
    id: str
    email: Optional[str] = None
    name: str
    bio: Optional[str] = None
    company: Optional[str] = None
    avatar: Optional[str] = None
    verified: Optional[bool] = None
    role: Optional[str] = None

    def __repr__(self):
        return (
            f"User( id: {self.id}, name: {self.name}, email: {self.email}, company:"
            f" {self.company} )"
        )

    def __str__(self) -> str:
        return self.__repr__()


class ResourceCollection(GraphQLBaseModel, Generic[T]):
    total_count: int
    items: List[T]
    cursor: Optional[str] = None


class ServerMigration(GraphQLBaseModel):
    moved_from: Optional[str]
    moved_to: Optional[str]


class AuthStrategy(GraphQLBaseModel):
    color: Optional[str]
    icon: str
    id: str
    name: str
    url: str


class ServerConfiguration(GraphQLBaseModel):
    blob_size_limit_bytes: int
    object_multipart_upload_size_limit_bytes: int
    object_size_limit_bytes: int


class ServerWorkspacesInfo(GraphQLBaseModel):
    workspaces_enabled: bool


# Keeping this one all Optionals at the minute,
#  because its used both as a deserialization model for GQL and Account Management
class ServerInfo(GraphQLBaseModel):
    name: Optional[str] = None
    company: Optional[str] = None
    url: Optional[str] = None
    admin_contact: Optional[str] = None
    description: Optional[str] = None
    canonical_url: Optional[str] = None
    scopes: Optional[List[dict]] = None
    auth_strategies: Optional[List[dict]] = None
    version: Optional[str] = None
    migration: Optional[ServerMigration] = None
    workspaces: Optional[ServerWorkspacesInfo] = None
    # TODO separate gql model from account management model


class LimitedUser(GraphQLBaseModel):
    """Limited user type, for showing public info about a user to another user."""

    id: str
    name: str
    bio: Optional[str]
    company: Optional[str]
    avatar: Optional[str]
    verified: Optional[bool]
    role: Optional[str]

    def __repr__(self):
        return (
            f"(name: {self.name}, "
            f"id: {self.id}, "
            f"bio: {self.bio}, "
            f"company: {self.company}, "
            f"verified: {self.verified}, "
            f"role: {self.role})"
        )


class PendingStreamCollaborator(GraphQLBaseModel):
    id: str
    invite_id: str
    stream_id: Optional[str] = None
    projectId: str
    stream_name: Optional[str] = None
    project_name: str
    title: str
    role: str
    invited_by: LimitedUser
    user: Optional[LimitedUser] = None
    token: Optional[str]

    def __repr__(self):
        return (
            f"PendingStreamCollaborator( inviteId: {self.invite_id}, streamId:"
            f" {self.stream_id}, role: {self.role}, title: {self.title}, invitedBy:"
            f" {self.user.name if self.user else None})"
        )

    def __str__(self) -> str:
        return self.__repr__()


class ProjectCollaborator(GraphQLBaseModel):
    id: str
    role: str
    user: LimitedUser


class Version(GraphQLBaseModel):
    author_user: Optional[LimitedUser]
    created_at: datetime
    id: str
    message: Optional[str]
    preview_url: str
    referenced_object: Optional[str]
    """Maybe null if workspaces version history limit has been exceeded"""
    source_application: Optional[str]


class Model(GraphQLBaseModel):
    author: Optional[LimitedUser]
    created_at: datetime
    description: Optional[str]
    display_name: str
    id: str
    name: str
    preview_url: Optional[str]
    updated_at: datetime


class ModelWithVersions(Model):
    versions: ResourceCollection[Version]


class ProjectPermissionChecks(GraphQLBaseModel):
    can_create_model: "PermissionCheckResult"
    can_delete: "PermissionCheckResult"
    can_load: "PermissionCheckResult"
    can_publish: "PermissionCheckResult"


class Project(GraphQLBaseModel):
    allow_public_comments: bool
    created_at: datetime
    description: Optional[str]
    id: str
    name: str
    role: Optional[str]
    source_apps: List[str]
    updated_at: datetime
    visibility: ProjectVisibility
    workspace_id: Optional[str]


class ProjectWithModels(Project):
    models: ResourceCollection[Model]


class ProjectWithPermissions(Project):
    permissions: ProjectPermissionChecks


class ProjectWithTeam(Project):
    invited_team: List[PendingStreamCollaborator]
    team: List[ProjectCollaborator]


class ProjectCommentCollection(ResourceCollection[T], Generic[T]):
    total_archived_count: int


class UserSearchResultCollection(GraphQLBaseModel):
    items: List[LimitedUser]
    cursor: Optional[str] = None


class PermissionCheckResult(GraphQLBaseModel):
    authorized: bool
    code: str
    message: str

    def ensure_authorised(self) -> None:
        """Raises WorkspacePermissionException if not authorized"""
        if not self.authorized:
            raise WorkspacePermissionException(self.message)


class WorkspacePermissionChecks(GraphQLBaseModel):
    can_create_project: PermissionCheckResult


class WorkspaceCreationState(GraphQLBaseModel):
    completed: bool


class LimitedWorkspace(GraphQLBaseModel):
    id: str
    name: str
    role: Optional[str]
    slug: str
    logo: Optional[str]
    description: Optional[str]


class Workspace(LimitedWorkspace):
    created_at: datetime
    updated_at: datetime
    read_only: bool
    creation_state: Optional[WorkspaceCreationState]
    permissions: WorkspacePermissionChecks



from typing import List, Optional

from gql import gql

from specklepy.core.api.inputs.user_inputs import (
    UserProjectsFilter,
    UserUpdateInput,
    UserWorkspacesFilter,
)
from specklepy.core.api.models import (
    PendingStreamCollaborator,
    Project,
    ResourceCollection,
    User,
)
from specklepy.core.api.models.current import (
    LimitedWorkspace,
    PermissionCheckResult,
    ProjectWithPermissions,
    Workspace,
)
from specklepy.core.api.resource import ResourceBase
from specklepy.core.api.responses import DataResponse
from specklepy.logging.exceptions import GraphQLException

NAME = "active_user"


class ActiveUserResource(ResourceBase):
    """API Access class for the active user"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )
        self.schema = User

    def get(self) -> Optional[User]:
        """Gets the currently active user profile
        (as extracted from the authorization header)

        Returns:
            User -- the requested user, or none if no authentication token
            is provided to the Client
        """
        QUERY = gql(
            """
            query User {
             data:activeUser {
               id
               email
               name
               bio
               company
               avatar
               verified
               role
             }
           }
           """
        )

        variables = {}

        return self.make_request_and_parse_response(
            DataResponse[Optional[User]], QUERY, variables
        ).data

    def update(self, input: UserUpdateInput) -> User:
        QUERY = gql(
            """
            mutation ActiveUserMutations($input: UserUpdateInput!) {
              data:activeUserMutations {
                data:update(user: $input) {
                  id
                  email
                  name
                  bio
                  company
                  avatar
                  verified
                  role
                }
              }
            }
           """
        )

        variables = {"input": input.model_dump(warnings="error", by_alias=True)}

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[User]], QUERY, variables
        ).data.data

    def get_projects(
        self,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        filter: Optional[UserProjectsFilter] = None,
    ) -> ResourceCollection[Project]:
        QUERY = gql(
            """
             query User($limit : Int!, $cursor: String, $filter: UserProjectsFilter) {
              data:activeUser {
                data:projects(limit: $limit, cursor: $cursor, filter: $filter) {
                   totalCount
                   cursor
                   items {
                      id
                      name
                      description
                      visibility
                      allowPublicComments
                      role
                      createdAt
                      updatedAt
                      sourceApps
                      workspaceId
                   }
                }
              }
            }
            """
        )

        variables = {
            "limit": limit,
            "cursor": cursor,
            "filter": filter.model_dump(warnings="error", by_alias=True)
            if filter
            else None,
        }

        response = self.make_request_and_parse_response(
            DataResponse[Optional[DataResponse[ResourceCollection[Project]]]],
            QUERY,
            variables,
        )

        if response.data is None:
            raise GraphQLException(
                "GraphQL response indicated that the ActiveUser could not be found"
            )

        return response.data.data

    def get_project_invites(self) -> List[PendingStreamCollaborator]:
        QUERY = gql(
            """
            query ProjectInvites {
              data:activeUser {
                data:projectInvites {
                  id
                  inviteId
                  invitedBy {
                    avatar
                    bio
                    company
                    id
                    name
                    role
                    verified
                  }
                  projectId
                  projectName
                  role
                  title
                  token
                  user {
                    id
                    name
                    bio
                    company
                    verified
                    avatar
                    role
                  }
                }
              }
            }
            """
        )

        variables = {}

        response = self.make_request_and_parse_response(
            DataResponse[Optional[DataResponse[List[PendingStreamCollaborator]]]],
            QUERY,
            variables,
        )

        if response.data is None:
            raise GraphQLException(
                "GraphQL response indicated that the ActiveUser could not be found"
            )

        return response.data.data

    def can_create_personal_projects(self) -> PermissionCheckResult:
        QUERY = gql(
            """
            query CanCreatePersonalProject {
              data:activeUser {
                data:permissions {
                  data:canCreatePersonalProject {
                    authorized
                    code
                    message
                  }
                }
              }
            }
            """
        )

        response = self.make_request_and_parse_response(
            DataResponse[Optional[DataResponse[DataResponse[PermissionCheckResult]]]],
            QUERY,
        )

        if response.data is None:
            raise GraphQLException(
                "GraphQL response indicated that the ActiveUser could not be found"
            )

        return response.data.data.data

    def get_workspaces(
        self,
        limit: int = 25,
        cursor: Optional[str] = None,
        filter: Optional[UserWorkspacesFilter] = None,
    ) -> ResourceCollection[Workspace]:
        """
        This feature is only available on Workspace enabled servers  (server versions
        >=2.23.17) e.g. app.speckle.systems
        """
        QUERY = gql(
            """
            query ActiveUser($limit: Int!, $cursor: String, $filter: UserWorkspacesFilter) {
              data:activeUser {
                data:workspaces(limit: $limit, cursor: $cursor, filter: $filter) {
                  cursor
                  totalCount
                  items {
                    id
                    name
                    role
                    slug
                    logo
                    createdAt
                    updatedAt
                    readOnly
                    description
                    creationState
                    {
                      completed
                    }
                    permissions {
                      canCreateProject {
                        authorized
                        code
                        message
                      }
                    }
                  }
                }
              }
            }
            """  # noqa: E501
        )

        variables = {
            "limit": limit,
            "cursor": cursor,
            "filter": filter.model_dump(warnings="error", by_alias=True)
            if filter
            else None,
        }

        response = self.make_request_and_parse_response(
            DataResponse[Optional[DataResponse[ResourceCollection[Workspace]]]],
            QUERY,
            variables,
        )

        if response.data is None:
            raise GraphQLException(
                "GraphQL response indicated that the ActiveUser could not be found"
            )

        return response.data.data

    def get_active_workspace(self) -> Optional[LimitedWorkspace]:
        """
        This feature is only available on Workspace enabled servers  (server versions
        >=2.23.17) e.g. app.speckle.systems
        """
        QUERY = gql(
            """
            query ActiveUser {
              data:activeUser {
                data:activeWorkspace {
                  id
                  name
                  role
                  slug
                  logo
                  description
                }
              }
            }
            """
        )

        response = self.make_request_and_parse_response(
            DataResponse[Optional[DataResponse[Optional[LimitedWorkspace]]]],
            QUERY,
        )

        if response.data is None:
            raise GraphQLException(
                "GraphQL response indicated that the ActiveUser could not be found"
            )

        return response.data.data

    def get_projects_with_permissions(
        self,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        filter: Optional[UserProjectsFilter] = None,
    ) -> ResourceCollection[ProjectWithPermissions]:
        """
        Gets the currently active user's projects with their permissions.
        This is useful for checking what actions can be performed on each project.
        """
        QUERY = gql(
            """
          query User($limit : Int!, $cursor: String, $filter: UserProjectsFilter) {
            data:activeUser {
              data:projects(limit: $limit, cursor: $cursor, filter: $filter) {
                totalCount
                cursor
                items {
                    id
                    name
                    description
                    visibility
                    allowPublicComments
                    role
                    createdAt
                    updatedAt
                    sourceApps
                    workspaceId
                    permissions {
                      canCreateModel {
                        code
                        authorized
                        message
                      }
                      canDelete {
                        code
                        authorized
                        message
                      }
                      canLoad {
                        code
                        authorized
                        message
                      }
                      canPublish {
                        code
                        authorized
                        message
                      }
                    }
                }
              }
            }
          }
          """
        )

        variables = {
            "limit": limit,
            "cursor": cursor,
            "filter": filter.model_dump(warnings="error", by_alias=True)
            if filter
            else None,
        }

        response = self.make_request_and_parse_response(
            DataResponse[
                Optional[DataResponse[ResourceCollection[ProjectWithPermissions]]]
            ],
            QUERY,
            variables,
        )

        if response.data is None:
            raise GraphQLException(
                "GraphQL response indicated that the ActiveUser could not be found"
            )

        return response.data.data



from typing import Optional

from gql import gql

from specklepy.core.api.models import (
    LimitedUser,
    UserSearchResultCollection,
)
from specklepy.core.api.resource import ResourceBase
from specklepy.core.api.responses import DataResponse

NAME = "other_user"


class OtherUserResource(ResourceBase):
    """API Access class for other users, that are not the currently active user."""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )
        self.schema = LimitedUser

    def get(self, id: str) -> Optional[LimitedUser]:
        """
        Gets the profile of another user.

        Arguments:
            id {str} -- the user id

        Returns:
            LimitedUser -- the retrieved profile of another user
        """
        QUERY = gql(
            """
          query LimitedUser($id: String!) {
            data:otherUser(id: $id){
              id
              name
              bio
              company
              avatar
              verified
              role
            }
          }
          """
        )

        variables = {"id": id}

        return self.make_request_and_parse_response(
            DataResponse[Optional[LimitedUser]], QUERY, variables
        ).data

    def user_search(
        self,
        query: str,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        archived: bool = False,
        emailOnly: bool = False,
    ) -> UserSearchResultCollection:
        """
        Searches for a user on the server, by name or email.
        The search query must be at least
        3 characters long

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
            cursor {Optional[str]} --
            archived {bool} --
            emailOnly {bool} --
        Returns:
            ResourceCollection[LimitedUser] -- User objects that match the search query
        """

        QUERY = gql(
            """
            query UserSearch(
                $query: String!,
                $limit: Int!,
                $cursor: String,
                $archived: Boolean,
                $emailOnly: Boolean
            ) {
              data:userSearch(
                query: $query,
                limit: $limit,
                cursor: $cursor,
                archived: $archived,
                emailOnly: $emailOnly
                ) {
                cursor
                items {
                 id
                 name
                 bio
                 company
                 avatar
                 verified
                 role
               }
             }
            }
            """
        )
        variables = {
            "query": query,
            "limit": limit,
            "cursor": cursor,
            "archived": archived,
            "emailOnly": emailOnly,
        }

        return self.make_request_and_parse_response(
            DataResponse[UserSearchResultCollection], QUERY, variables
        ).data



from dataclasses import dataclass
from typing import List

from specklepy.objects.base import Base
from specklepy.objects.geometry.point import Point
from specklepy.objects.interfaces import IHasUnits


@dataclass(kw_only=True)
class PointCloud(Base, IHasUnits, speckle_type="Objects.Geometry.PointCloud"):
    """
    a collection of 3-dimensional points
    """

    points: List[Point]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"points: {len(self.points)}, "
            f"units: {self.units})"
        )

    # sizes and colors could be added in the future



from dataclasses import dataclass

from specklepy.objects.geometry.point import Point


@dataclass(kw_only=True)
class ControlPoint(Point, speckle_type="Objects.Geometry.ControlPoint"):
    """
    a single 3-dimensional point with weight
    """

    weight: float

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"x: {self.x}, "
            f"y: {self.y}, "
            f"z: {self.z}, "
            f"weight: {self.weight}, "
            f"units: {self.units})"
        )



from typing import Optional

from specklepy.core.api.enums import (
    ProjectModelsUpdatedMessageType,
    ProjectUpdatedMessageType,
    ProjectVersionsUpdatedMessageType,
    UserProjectsUpdatedMessageType,
)
from specklepy.core.api.models.current import Model, Project, Version
from specklepy.core.api.models.graphql_base_model import GraphQLBaseModel


class UserProjectsUpdatedMessage(GraphQLBaseModel):
    id: str
    type: UserProjectsUpdatedMessageType
    project: Optional[Project]


class ProjectModelsUpdatedMessage(GraphQLBaseModel):
    id: str
    type: ProjectModelsUpdatedMessageType
    model: Optional[Model]


class ProjectUpdatedMessage(GraphQLBaseModel):
    id: str
    type: ProjectUpdatedMessageType
    project: Optional[Project]


class ProjectVersionsUpdatedMessage(GraphQLBaseModel):
    id: str
    type: ProjectVersionsUpdatedMessageType
    model_id: str
    version: Optional[Version]
